from dataclasses import dataclass, field, fields
from typing import List, Optional, Tuple, Union

from qibolab.channels import Channel
from qibolab.native import SingleQubitNatives, TwoQubitNatives

QubitId = Union[str, int]
"""Type for qubit names."""

CHANNEL_NAMES = ("readout", "feedback", "drive", "flux", "twpa")
"""Names of channels that belong to a qubit.
Not all channels are required to operate a qubit.
"""
EXCLUDED_FIELDS = CHANNEL_NAMES + ("name", "flux_coupler", "native_gates")
"""Qubit dataclass fields that are excluded by the ``characterization`` property."""


@dataclass
class Qubit:
    """Representation of a physical qubit.

    Qubit objects are instantiated by :class:`qibolab.platforms.platform.Platform`
    but they are passed to instrument designs in order to play pulses.

    Args:
        name (int, str): Qubit number or name.
        readout (:class:`qibolab.platforms.utils.Channel`): Channel used to
            readout pulses to the qubit.
        feedback (:class:`qibolab.platforms.utils.Channel`): Channel used to
            get readout feedback from the qubit.
        drive (:class:`qibolab.platforms.utils.Channel`): Channel used to
            send drive pulses to the qubit.
        flux (:class:`qibolab.platforms.utils.Channel`): Channel used to
            send flux pulses to the qubit.
        Other characterization parameters for the qubit, loaded from the runcard.
    """

    name: QubitId

    bare_resonator_frequency: int = 0
    readout_frequency: int = 0  # this is the dressed frequency
    drive_frequency: int = 0
    anharmonicity: int = 0
    Ec: int = 0
    Ej: int = 0
    g: int = 0
    assigment_fidelity: float = 0.0
    sweetspot: float = 0
    peak_voltage: float = 0
    pi_pulse_amplitude: float = 0
    T1: int = 0
    T2: int = 0
    T2_spin_echo: int = 0
    state0_voltage: int = 0
    state1_voltage: int = 0
    mean_gnd_states: List[float] = field(default_factory=lambda: [0, 0])
    mean_exc_states: List[float] = field(default_factory=lambda: [0, 0])
    resonator_polycoef_flux: List[float] = field(default_factory=list)

    # parameters for single shot classification
    threshold: Optional[float] = None
    iq_angle: float = 0.0
    # required for mixers (not sure if it should be here)
    mixer_drive_g: float = 0.0
    mixer_drive_phi: float = 0.0
    mixer_readout_g: float = 0.0
    mixer_readout_phi: float = 0.0

    readout: Optional[Channel] = None
    feedback: Optional[Channel] = None
    twpa: Optional[Channel] = None
    drive: Optional[Channel] = None
    flux: Optional[Channel] = None
    flux_coupler: Optional[List["Qubit"]] = None

    classifiers_hpars: dict = field(default_factory=dict)
    native_gates: SingleQubitNatives = field(default_factory=SingleQubitNatives)

    def __post_init__(self):
        # register qubit in ``flux`` channel so that we can access
        # ``sweetspot`` and ``filters`` at the channel level
        if self.flux:
            self.flux.qubit = self

    @property
    def channels(self):
        for name in CHANNEL_NAMES:
            channel = getattr(self, name)
            if channel is not None:
                yield channel

    @property
    def characterization(self):
        """Dictionary containing characterization parameters."""
        return {fld.name: getattr(self, fld.name) for fld in fields(self) if fld.name not in EXCLUDED_FIELDS}


QubitPairId = Tuple[QubitId, QubitId]
"""Type for holding ``QubitPair``s in the ``platform.pairs`` dictionary."""


@dataclass
class QubitPair:
    """Data structure for holding the native two-qubit gates acting on a pair of qubits.

    This is needed for symmetry to the single-qubit gates which are storred in the
    :class:`qibolab.platforms.abstract.Qubit`.

    Qubits are sorted according to ``qubit.name`` such that
    ``qubit1.name < qubit2.name``.
    """

    qubit1: Qubit
    qubit2: Qubit
    native_gates: TwoQubitNatives = field(default_factory=TwoQubitNatives)



import numpy as np


def print_readable_snapshot(device, file, update: bool = False, max_chars: int = 80) -> None:
    """
    Prints a readable version of the snapshot.
    The readable snapshot includes the name, value and unit of each
    parameter.
    A convenience function to quickly get an overview of the
    status of an instrument.

    Args:
        update: If ``True``, update the state by querying the
            instrument. If ``False``, just use the latest values in memory.
            This argument gets passed to the snapshot function.
        max_chars: the maximum number of characters per line. The
            readable snapshot will be cropped if this value is exceeded.
            Defaults to 80 to be consistent with default terminal width.
    """
    floating_types = (float, np.integer, np.floating)
    snapshot = device.snapshot(update=update)

    par_lengths = [len(p) for p in snapshot["parameters"]]

    # Min of 50 is to prevent a super long parameter name to break this
    # function
    par_field_len = min(max(par_lengths) + 1, 50)

    file.write(device.name + ":" + "\n")
    file.write("{0:<{1}}".format("\tparameter ", par_field_len) + "value" + "\n")
    file.write("-" * max_chars + "\n")
    for par in sorted(snapshot["parameters"]):
        name = snapshot["parameters"][par]["name"]
        msg = "{0:<{1}}:".format(name, par_field_len)

        # in case of e.g. ArrayParameters, that usually have
        # snapshot_value == False, the parameter may not have
        # a value in the snapshot
        val = snapshot["parameters"][par].get("value", "Not available")

        unit = snapshot["parameters"][par].get("unit", None)
        if unit is None:
            # this may be a multi parameter
            unit = snapshot["parameters"][par].get("units", None)
        if isinstance(val, floating_types):
            msg += f"\t{val:.5g} "
            # numpy float and int types format like builtins
        else:
            msg += f"\t{val} "
        if unit != "":  # corresponds to no unit
            msg += f"({unit})"
        # Truncate the message if it is longer than max length
        if len(msg) > max_chars and not max_chars == -1:
            msg = msg[0 : max_chars - 3] + "..."
        file.write(msg + "\n")

    for submodule in device.submodules.values():
        print_readable_snapshot(submodule, file, update=update, max_chars=max_chars)



import signal

import numpy as np
from qibo.config import log, raise_error

from qibolab import AcquisitionType, AveragingMode, ExecutionParameters
from qibolab.instruments.abstract import Controller
from qibolab.instruments.qblox.cluster import Cluster
from qibolab.instruments.qblox.cluster_qcm_bb import ClusterQCM_BB
from qibolab.instruments.qblox.cluster_qcm_rf import ClusterQCM_RF
from qibolab.instruments.qblox.cluster_qrm_rf import ClusterQRM_RF
from qibolab.pulses import PulseSequence, PulseType
from qibolab.sweeper import Parameter, Sweeper, SweeperType


class QbloxController(Controller):
    """A controller to manage qblox devices.

    Attributes:
        is_connected (bool): .
        modules (dict): A dictionay with the qblox modules connected to the experiment.
    """

    def __init__(self, name, cluster, modules):
        """Initialises the controller."""
        super().__init__(name=name, address="")
        self.is_connected = False
        self.cluster: Cluster = cluster
        self.modules: dict = modules
        signal.signal(signal.SIGTERM, self._termination_handler)

    def connect(self):
        """Connects to the modules."""

        if not self.is_connected:
            try:
                self.cluster.connect()
                for name in self.modules:
                    self.modules[name].connect(self.cluster.device)
                self.is_connected = True
            except Exception as exception:
                raise_error(
                    RuntimeError,
                    "Cannot establish connection to " f"{self.modules[name]} module. " f"Error captured: '{exception}'",
                )
                # TODO: check for exception 'The module qrm_rf0 does not have parameters in0_att' and reboot the cluster

            else:
                log.info(f"QbloxController: all modules connected.")

    def setup(self):
        """Sets all modules up."""

        if not self.is_connected:
            raise_error(
                RuntimeError,
                "There is no connection to the modules, the setup cannot be completed",
            )
        self.cluster.setup()
        for name in self.modules:
            self.modules[name].setup()

    def start(self):
        """Starts all modules."""
        self.cluster.start()
        if self.is_connected:
            for name in self.modules:
                self.modules[name].start()

    def stop(self):
        """Stops all modules."""

        if self.is_connected:
            for name in self.modules:
                self.modules[name].stop()
        self.cluster.stop()

    def _termination_handler(self, signum, frame):
        """Calls all modules to stop if the program receives a termination signal."""

        log.warning("Termination signal received, stopping modules.")
        if self.is_connected:
            for name in self.modules:
                self.modules[name].stop()
        log.warning("QbloxController: all modules stopped.")
        exit(0)

    def disconnect(self):
        """Disconnects all modules."""

        if self.is_connected:
            for name in self.modules:
                self.modules[name].disconnect()
            self.cluster.disconnect()
            self.is_connected = False

    def _execute_pulse_sequence(
        self,
        qubits: dict,
        sequence: PulseSequence,
        options: ExecutionParameters,
        sweepers: list() = [],  # list(Sweeper) = []
        **kwargs
        # nshots=None,
        # navgs=None,
        # relaxation_time=None,
    ):
        """Executes a sequence of pulses or a sweep.

        Args:
            sequence (:class:`qibolab.pulses.PulseSequence`): The sequence of pulses to execute.
            options (:class:`qibolab.platforms.platform.ExecutionParameters`): Object holding the execution options.
            sweepers (list(Sweeper)): A list of Sweeper objects defining parameter sweeps.
        """
        if not self.is_connected:
            raise_error(RuntimeError, "Execution failed because modules are not connected.")

        if options.averaging_mode == AveragingMode.SINGLESHOT:
            nshots = options.nshots
            navgs = 1
            average = False
        else:
            navgs = options.nshots
            nshots = 1
            average = True

        relaxation_time = options.relaxation_time
        repetition_duration = sequence.finish + relaxation_time

        # shots results are stored in separate bins
        # calculate number of shots
        num_bins = nshots
        for sweeper in sweepers:
            num_bins *= len(sweeper.values)

        # DEBUG: Plot Pulse Sequence
        # sequence.plot('plot.png')
        # DEBUG: sync_en
        # from qblox_instruments.qcodes_drivers.cluster import Cluster
        # cluster:Cluster = self.modules['cluster'].device
        # for module in cluster.modules:
        #     if module.get("present"):
        #         for sequencer in module.sequencers:
        #             if sequencer.get('sync_en'):
        #                 print(f"type: {module.module_type}, sequencer: {sequencer.name}, sync_en: True")

        # Process Pulse Sequence. Assign pulses to modules and generate waveforms & program
        module_pulses = {}
        data = {}
        for name in self.modules:
            # from the pulse sequence, select those pulses to be synthesised by the module
            module_pulses[name] = sequence.get_channel_pulses(*self.modules[name].channels)

            for port in self.modules[name].ports:
                # _los = []
                # _ifs = []
                port_pulses = module_pulses[name].get_channel_pulses(self.modules[name]._port_channel_map[port])
                # for pulse in port_pulses:
                #     if pulse.type == PulseType.READOUT:
                #         _if = int(self.native_gates["single_qubit"][pulse.qubit]["MZ"]["if_frequency"])
                #         pulse._if = _if
                #         _los.append(int(pulse.frequency - _if))
                #         _ifs.append(int(_if))
                #     elif pulse.type == PulseType.DRIVE:
                #         _if = int(self.native_gates["single_qubit"][pulse.qubit]["RX"]["if_frequency"])
                #         pulse._if = _if
                #         _los.append(int(pulse.frequency - _if))
                #         _ifs.append(int(_if))

                # # where multiple qubits share the same lo (for example on a readout line), check lo consistency
                # if len(_los) > 1:
                #     for _ in range(1, len(_los)):
                #         if _los[0] != _los[_]:
                #             raise ValueError(
                #                 f"""Pulses:
                #                 {module_pulses[name]}
                #                 sharing the lo at device: {name} - port: {port}
                #                 cannot be synthesised with intermediate frequencies:
                #                 {_ifs}"""
                #             )
                # if len(_los) > 0:
                #     self.modules[name].ports[port].lo_frequency = _los[0]

                if isinstance(self.modules[name], (ClusterQRM_RF, ClusterQCM_RF)):
                    # without access to _ifs _los cannot be set ^^^^^^
                    for pulse in port_pulses:
                        pulse._if = int(pulse.frequency - self.modules[name].ports[port].lo_frequency)

            #  ask each module to generate waveforms & program and upload them to the device
            self.modules[name].process_pulse_sequence(
                qubits, module_pulses[name], navgs, nshots, repetition_duration, sweepers
            )

            # log.info(f"{self.modules[name]}: Uploading pulse sequence")
            self.modules[name].upload()

        # play the sequence or sweep
        for name in self.modules:
            if isinstance(self.modules[name], (ClusterQRM_RF, ClusterQCM_RF, ClusterQCM_BB)):
                self.modules[name].play_sequence()

        # retrieve the results
        acquisition_results = {}
        for name in self.modules:
            if isinstance(self.modules[name], ClusterQRM_RF):
                if not module_pulses[name].ro_pulses.is_empty:
                    results = self.modules[name].acquire()
                    existing_keys = set(acquisition_results.keys()) & set(results.keys())
                    for key, value in results.items():
                        if key in existing_keys:
                            acquisition_results[key].update(value)
                        else:
                            acquisition_results[key] = value

        # TODO: move to QRM_RF.acquire()
        for ro_pulse in sequence.ro_pulses:
            if options.acquisition_type is AcquisitionType.DISCRIMINATION:
                _res = acquisition_results[ro_pulse.serial][2]
                if average:
                    _res = np.mean(_res, axis=0)
            else:
                ires = acquisition_results[ro_pulse.serial][0]
                qres = acquisition_results[ro_pulse.serial][1]
                _res = ires + 1j * qres

            acquisition = options.results_type(_res)
            data[ro_pulse.serial] = data[ro_pulse.qubit] = acquisition

            # data[ro_pulse.serial] = ExecutionResults.from_components(*acquisition_results[ro_pulse.serial])
            # data[ro_pulse.serial] = IntegratedResults(acquisition_results[ro_pulse.serial])
            # data[ro_pulse.qubit] = copy.copy(data[ro_pulse.serial])
        return data

    def play(self, qubits, sequence, options):
        return self._execute_pulse_sequence(qubits, sequence, options)

    def play_sequences(self, *args, **kwargs):
        raise_error(NotImplementedError, "play_sequences is not implemented in qblox driver yet.")

    def sweep(self, qubits: dict, sequence: PulseSequence, options: ExecutionParameters, *sweepers):
        """Executes a sequence of pulses while sweeping one or more parameters.

        The parameters to be swept are defined in :class:`qibolab.sweeper.Sweeper` object.
        Args:
            sequence (:class:`qibolab.pulses.PulseSequence`): The sequence of pulses to execute.
            options (:class:`qibolab.platforms.platform.ExecutionParameters`): Object holding the execution options.
            sweepers (list(Sweeper)): A list of Sweeper objects defining parameter sweeps.
        """
        id_results = {}
        map_id_serial = {}

        # during the sweep, pulse parameters need to be changed
        # to avoid affecting the user, make a copy of the pulse sequence
        # and the sweepers, as they contain references to pulses
        sequence_copy = sequence.copy()
        sweepers_copy = []
        for sweeper in sweepers:
            if sweeper.pulses:
                ps = []
                for pulse in sweeper.pulses:
                    if pulse in sequence_copy:
                        ps.append(sequence_copy[sequence_copy.index(pulse)])
            else:
                ps = None
            sweepers_copy.append(
                Sweeper(
                    parameter=sweeper.parameter,
                    values=sweeper.values,
                    pulses=ps,
                    qubits=sweeper.qubits,
                    type=sweeper.type,
                )
            )

        # reverse sweepers exept for res punchout att
        contains_attenuation_frequency = any(
            sweepers_copy[i].parameter == Parameter.attenuation
            and sweepers_copy[i + 1].parameter == Parameter.frequency
            for i in range(len(sweepers_copy) - 1)
        )

        if not contains_attenuation_frequency:
            sweepers_copy.reverse()

        # create a map between the pulse id, which never changes, and the original serial
        for pulse in sequence_copy.ro_pulses:
            map_id_serial[pulse.id] = pulse.serial
            id_results[pulse.id] = None
            id_results[pulse.qubit] = None

        # execute the each sweeper recursively
        self._sweep_recursion(
            qubits,
            sequence_copy,
            options,
            *tuple(sweepers_copy),
            results=id_results,
        )

        # return the results using the original serials
        serial_results = {}
        for pulse in sequence_copy.ro_pulses:
            serial_results[map_id_serial[pulse.id]] = id_results[pulse.id]
            serial_results[pulse.qubit] = id_results[pulse.id]
        return serial_results

    def _sweep_recursion(
        self,
        qubits,
        sequence,
        options: ExecutionParameters,
        *sweepers,
        results,
    ):
        """Executes a sweep recursively.

        Args:
            sequence (:class:`qibolab.pulses.PulseSequence`): The sequence of pulses to execute.
            sweepers (list(Sweeper)): A list of Sweeper objects defining parameter sweeps.
            results (:class:`qibolab.results.ExecutionResults`): A results object to update with the reults of the execution.
            nshots (int): The number of times the sequence of pulses should be executed.
            average (bool): A flag to indicate if the results of the shots should be averaged.
            relaxation_time (int): The the time to wait between repetitions to allow the qubit relax to ground state.
        """

        for_loop_sweepers = [Parameter.attenuation, Parameter.lo_frequency]
        rt_sweepers = [
            Parameter.frequency,
            Parameter.gain,
            Parameter.bias,
            Parameter.amplitude,
            Parameter.start,
            Parameter.duration,
            Parameter.relative_phase,
        ]

        sweeper: Sweeper = sweepers[0]

        # until sweeper contains the information to determine whether the sweep should be relative or
        # absolute:

        # elif sweeper.parameter is Parameter.relative_phase:
        #     initial = {}
        #     for pulse in sweeper.pulses:
        #         initial[pulse.id] = pulse.relative_phase

        # elif sweeper.parameter is Parameter.frequency:
        #     initial = {}
        #     for pulse in sweeper.pulses:
        #         initial[pulse.id] = pulse.frequency

        if sweeper.parameter in for_loop_sweepers:
            # perform sweep recursively
            for value in sweeper.values:
                if sweeper.parameter is Parameter.attenuation:
                    initial = {}
                    for qubit in sweeper.qubits:
                        initial[qubit.name] = qubits[qubit.name].readout.attenuation
                        if sweeper.type == SweeperType.ABSOLUTE:
                            qubit.readout.attenuation = value
                        elif sweeper.type == SweeperType.OFFSET:
                            qubit.readout.attenuation = initial[qubit.name] + value
                        elif sweeper.type == SweeperType.FACTOR:
                            qubit.readout.attenuation = initial[qubit.name] * value

                elif sweeper.parameter is Parameter.lo_frequency:
                    initial = {}
                    for pulse in sweeper.pulses:
                        if pulse.type == PulseType.READOUT:
                            initial[pulse.id] = qubits[pulse.qubit].readout.lo_frequency
                            if sweeper.type == SweeperType.ABSOLUTE:
                                qubits[pulse.qubit].readout.lo_frequency = value
                            elif sweeper.type == SweeperType.OFFSET:
                                qubits[pulse.qubit].readout.lo_frequency = initial[pulse.id] + value
                            elif sweeper.type == SweeperType.FACTOR:
                                qubits[pulse.qubit].readout.lo_frequency = initial[pulse.id] * value

                        elif pulse.type == PulseType.DRIVE:
                            initial[pulse.id] = qubits[pulse.qubit].drive.lo_frequency
                            if sweeper.type == SweeperType.ABSOLUTE:
                                qubits[pulse.qubit].drive.lo_frequency = value
                            elif sweeper.type == SweeperType.OFFSET:
                                qubits[pulse.qubit].drive.lo_frequency = initial[pulse.id] + value
                            elif sweeper.type == SweeperType.FACTOR:
                                qubits[pulse.qubit].drive.lo_frequency = initial[pulse.id] * value

                if len(sweepers) > 1:
                    self._sweep_recursion(
                        qubits,
                        sequence,
                        options,
                        *sweepers[1:],
                        results=results,
                    )
                else:
                    result = self._execute_pulse_sequence(qubits=qubits, sequence=sequence, options=options)
                    for pulse in sequence.ro_pulses:
                        if results[pulse.id]:
                            results[pulse.id] += result[pulse.serial]
                        else:
                            results[pulse.id] = result[pulse.serial]
                        results[pulse.qubit] = results[pulse.id]
        else:
            # rt sweeps
            # relative phase sweeps that cross 0 need to be split in two separate sweeps
            split_relative_phase = False
            if sweeper.parameter == Parameter.relative_phase:
                if not sweeper.type == SweeperType.ABSOLUTE:
                    raise_error(ValueError, "relative_phase sweeps other than ABSOLUTE are not supported by qblox yet")
                from qibolab.instruments.qblox.q1asm import convert_phase

                c_values = np.array([convert_phase(v) for v in sweeper.values])
                if any(np.diff(c_values) < 0):
                    split_relative_phase = True
                    _from = 0
                    for idx in np.append(np.where(np.diff(c_values) < 0), len(c_values) - 1):
                        _to = idx + 1
                        _values = sweeper.values[_from:_to]
                        split_sweeper = Sweeper(
                            parameter=sweeper.parameter,
                            values=_values,
                            pulses=sweeper.pulses,
                            qubits=sweeper.qubits,
                        )
                        self._sweep_recursion(
                            qubits,
                            sequence,
                            options,
                            *(tuple([split_sweeper]) + sweepers[1:]),
                            results=results,
                        )
                        _from = _to

            if not split_relative_phase:
                if all(s.parameter in rt_sweepers for s in sweepers):
                    nshots = options.nshots if options.averaging_mode == AveragingMode.SINGLESHOT else 1
                    navgs = options.nshots if options.averaging_mode != AveragingMode.SINGLESHOT else 1
                    num_bins = nshots
                    for sweeper in sweepers:
                        num_bins *= len(sweeper.values)

                    # split the sweep if the number of bins is larget than the memory of the sequencer (2**17)
                    if num_bins < 2**17:
                        # for sweeper in sweepers:
                        #     if sweeper.parameter is Parameter.amplitude:
                        #         # qblox cannot sweep amplitude in real time, but sweeping gain is quivalent
                        #         for pulse in sweeper.pulses:
                        #             pulse.amplitude = 1

                        #     elif sweeper.parameter is Parameter.gain:
                        #         for pulse in sweeper.pulses:
                        #             # qblox has an external and an internal gains
                        #             # when sweeping the internal, set the external to 1
                        #             # TODO check if it needs to be restored after execution
                        #             if pulse.type == PulseType.READOUT:
                        #                 qubits[pulse.qubit].readout.gain = 1
                        #             elif pulse.type == PulseType.DRIVE:
                        #                 qubits[pulse.qubit].drive.gain = 1

                        result = self._execute_pulse_sequence(qubits, sequence, options, sweepers)
                        for pulse in sequence.ro_pulses:
                            if results[pulse.id]:
                                results[pulse.id] += result[pulse.serial]
                            else:
                                results[pulse.id] = result[pulse.serial]
                            results[pulse.qubit] = results[pulse.id]
                    else:
                        sweepers_repetitions = 1
                        for sweeper in sweepers:
                            sweepers_repetitions *= len(sweeper.values)
                        if sweepers_repetitions < 2**17:
                            # split nshots
                            max_rt_nshots = (2**17) // sweepers_repetitions
                            num_full_sft_iterations = nshots // max_rt_nshots
                            num_bins = max_rt_nshots * sweepers_repetitions

                            for sft_iteration in range(num_full_sft_iterations + 1):
                                _nshots = min(max_rt_nshots, nshots - sft_iteration * max_rt_nshots)
                                self._sweep_recursion(
                                    qubits,
                                    sequence,
                                    options,
                                    *sweepers,
                                    results=results,
                                )
                        else:
                            for shot in range(nshots):
                                num_bins = 1
                                for sweeper in sweepers[1:]:
                                    num_bins *= len(sweeper.values)
                                sweeper = sweepers[0]
                                max_rt_iterations = (2**17) // num_bins
                                num_full_sft_iterations = len(sweeper.values) // max_rt_iterations
                                num_bins = nshots * max_rt_iterations
                                for sft_iteration in range(num_full_sft_iterations + 1):
                                    _from = sft_iteration * max_rt_iterations
                                    _to = min((sft_iteration + 1) * max_rt_iterations, len(sweeper.values))
                                    _values = sweeper.values[_from:_to]
                                    split_sweeper = Sweeper(
                                        parameter=sweeper.parameter,
                                        values=_values,
                                        pulses=sweeper.pulses,
                                        qubits=sweeper.qubits,
                                    )

                                    self._sweep_recursion(
                                        qubits,
                                        sequence,
                                        options,
                                        *(tuple([split_sweeper]) + sweepers[1:]),
                                        results=results,
                                    )
                else:
                    # TODO: reorder the sequence of the sweepers and the results
                    raise Exception("cannot execute a for-loop sweeper nested inside of a rt sweeper")



import numpy as np
from qibo import gates, matrices
from qibo.config import raise_error
from scipy.linalg import expm

magic_basis = np.array([[1, -1j, 0, 0], [0, 0, 1, -1j], [0, 0, -1, -1j], [1, 1j, 0, 0]]) / np.sqrt(2)

bell_basis = np.array([[1, 1, 0, 0], [0, 0, 1, 1], [0, 0, 1, -1], [1, -1, 0, 0]]) / np.sqrt(2)

H = np.array([[1, 1], [1, -1]]) / np.sqrt(2)


def u3_decomposition(unitary):
    """Decomposes arbitrary one-qubit gates to U3.

    Args:
        unitary (np.ndarray): Unitary 2x2 matrix to be decomposed.

    Returns:
        theta, phi, lam: parameters of U3 gate.
    """
    # https://github.com/Qiskit/qiskit-terra/blob/d2e3340adb79719f9154b665e8f6d8dc26b3e0aa/qiskit/quantum_info/synthesis/one_qubit_decompose.py#L221
    su2 = unitary / np.sqrt(np.linalg.det(unitary))
    theta = 2 * np.arctan2(abs(su2[1, 0]), abs(su2[0, 0]))
    plus = np.angle(su2[1, 1])
    minus = np.angle(su2[1, 0])
    phi = plus + minus
    lam = plus - minus
    return theta, phi, lam


def calculate_psi(unitary):
    """Solves the eigenvalue problem of UT_U.

    See step (1) of Appendix A in arXiv:quant-ph/0011050.

    Args:
        unitary (np.ndarray): Unitary matrix of the gate we are
        decomposing in the computational basis.

    Returns:
        Eigenvectors (in the computational basis) and eigenvalues
        of UT_U.
    """
    # write unitary in magic basis
    u_magic = np.dot(np.dot(np.conj(magic_basis.T), unitary), magic_basis)
    # construct and diagonalize UT_U
    ut_u = np.dot(u_magic.T, u_magic)
    # When the matrix given to np.linalg.eig is a diagonal matrix up to machine precision the decomposition
    # is not accurate anymore. decimals = 20 works for random 2q Clifford unitaries.
    eigvals, psi_magic = np.linalg.eig(np.round(ut_u, decimals=20))
    # orthogonalize eigenvectors in the case of degeneracy (Gram-Schmidt)
    psi_magic, _ = np.linalg.qr(psi_magic)
    # write psi in computational basis
    psi = np.dot(magic_basis, psi_magic)
    return psi, eigvals


def schmidt_decompose(state):
    """Decomposes a two-qubit product state to its single-qubit parts."""
    u, d, v = np.linalg.svd(np.reshape(state, (2, 2)))
    if not np.allclose(d, [1, 0]):  # pragma: no cover
        raise_error(
            ValueError,
            f"Unexpected singular values: {d}\nCan only decompose product states.",
        )
    return u[:, 0], v[0]


def calculate_single_qubit_unitaries(psi):
    """Calculates local unitaries that maps a maximally entangled basis to the magic basis.

    See Lemma 1 of Appendix A in arXiv:quant-ph/0011050.

    Args:
        psi (np.ndarray): Maximally entangled two-qubit states that define a basis.

    Returns:
        Local unitaries UA and UB that map the given basis to the magic basis.
    """

    # TODO: Handle the case where psi is not real in the magic basis
    psi_magic = np.dot(np.conj(magic_basis).T, psi)
    if not np.allclose(psi_magic.imag, np.zeros_like(psi_magic)):  # pragma: no cover
        raise_error(NotImplementedError, "Given state is not real in the magic basis.")
    psi_bar = np.copy(psi).T

    # find e and f by inverting (A3), (A4)
    ef = (psi_bar[0] + 1j * psi_bar[1]) / np.sqrt(2)
    e_f_ = (psi_bar[0] - 1j * psi_bar[1]) / np.sqrt(2)
    e, f = schmidt_decompose(ef)
    e_, f_ = schmidt_decompose(e_f_)
    # find exp(1j * delta) using (A5a)
    ef_ = np.kron(e, f_)
    phase = 1j * np.sqrt(2) * np.dot(np.conj(ef_), psi_bar[2])

    # construct unitaries UA, UB using (A6a), (A6b)
    ua = np.tensordot([1, 0], np.conj(e), axes=0) + phase * np.tensordot([0, 1], np.conj(e_), axes=0)
    ub = np.tensordot([1, 0], np.conj(f), axes=0) + np.conj(phase) * np.tensordot([0, 1], np.conj(f_), axes=0)
    return ua, ub


def calculate_diagonal(unitary, ua, ub, va, vb):
    """Calculates Ud matrix that can be written as exp(-iH).

    See Eq. (A1) in arXiv:quant-ph/0011050.
    Ud is diagonal in the magic and Bell basis.
    """
    # normalize U_A, U_B, V_A, V_B so that detU_d = 1
    # this is required so that sum(lambdas) = 0
    # and Ud can be written as exp(-iH)
    det = np.linalg.det(unitary) ** (1 / 16)
    ua *= det
    ub *= det
    va *= det
    vb *= det
    u_dagger = np.conj(np.kron(ua, ub).T)
    v_dagger = np.conj(np.kron(va, vb).T)
    ud = np.dot(np.dot(u_dagger, unitary), v_dagger)
    return ua, ub, ud, va, vb


def magic_decomposition(unitary):
    """Decomposes an arbitrary unitary to (A1) from arXiv:quant-ph/0011050."""
    psi, eigvals = calculate_psi(unitary)
    psi_tilde = np.conj(np.sqrt(eigvals)) * np.dot(unitary, psi)
    va, vb = calculate_single_qubit_unitaries(psi)
    ua_dagger, ub_dagger = calculate_single_qubit_unitaries(psi_tilde)
    ua, ub = np.conj(ua_dagger.T), np.conj(ub_dagger.T)
    return calculate_diagonal(unitary, ua, ub, va, vb)


def to_bell_diagonal(ud):
    """Transforms a matrix to the Bell basis and checks if it is diagonal."""
    ud_bell = np.dot(np.dot(np.conj(bell_basis).T, ud), bell_basis)
    ud_diag = np.diag(ud_bell)
    if not np.allclose(np.diag(ud_diag), ud_bell):  # pragma: no cover
        return None
    uprod = np.prod(ud_diag)
    if not np.allclose(uprod, 1):  # pragma: no cover
        return None
    return ud_diag


def calculate_h_vector(ud_diag):
    """Finds h parameters corresponding to exp(-iH).

    See Eq. (4)-(5) in arXiv:quant-ph/0307177.
    """
    lambdas = -np.angle(ud_diag)
    hx = (lambdas[0] + lambdas[2]) / 2.0
    hy = (lambdas[1] + lambdas[2]) / 2.0
    hz = (lambdas[0] + lambdas[1]) / 2.0
    return hx, hy, hz


def cnot_decomposition(q0, q1, hx, hy, hz):
    """Performs decomposition (6) from arXiv:quant-ph/0307177."""
    u3 = -1j * (matrices.X + matrices.Z) / np.sqrt(2)
    # use corrected version from PRA paper (not arXiv)
    u2 = -u3 @ expm(-1j * (hx - np.pi / 4) * matrices.X)
    # add an extra exp(-i pi / 4) global phase to get exact match
    v2 = expm(-1j * hz * matrices.Z) * np.exp(-1j * np.pi / 4)
    v3 = expm(1j * hy * matrices.Z)
    w = (matrices.I - 1j * matrices.X) / np.sqrt(2)
    # change CNOT to CZ using Hadamard gates
    return [
        gates.H(q1),
        gates.CZ(q0, q1),
        gates.Unitary(u2, q0),
        gates.Unitary(H @ v2 @ H, q1),
        gates.CZ(q0, q1),
        gates.Unitary(u3, q0),
        gates.Unitary(H @ v3 @ H, q1),
        gates.CZ(q0, q1),
        gates.Unitary(w, q0),
        gates.Unitary(np.conj(w).T @ H, q1),
    ]


def cnot_decomposition_light(q0, q1, hx, hy):
    """Performs decomposition (24) from arXiv:quant-ph/0307177."""
    w = (matrices.I - 1j * matrices.X) / np.sqrt(2)
    u2 = expm(-1j * hx * matrices.X)
    v2 = expm(1j * hy * matrices.Z)
    # change CNOT to CZ using Hadamard gates
    return [
        gates.Unitary(np.conj(w).T, q0),
        gates.Unitary(H @ w, q1),
        gates.CZ(q0, q1),
        gates.Unitary(u2, q0),
        gates.Unitary(H @ v2 @ H, q1),
        gates.CZ(q0, q1),
        gates.Unitary(w, q0),
        gates.Unitary(np.conj(w).T @ H, q1),
    ]


def two_qubit_decomposition(q0, q1, unitary):
    """Performs two qubit unitary gate decomposition (24) from arXiv:quant-ph/0307177.

    Args:
        q0, q1 (int): Target qubits
        unitary (np.ndarray): Unitary 4x4 matrix we are decomposing.

    Returns:
        list of gates implementing decomposition (24) from arXiv:quant-ph/0307177
    """
    ud_diag = to_bell_diagonal(unitary)
    ud = None
    if ud_diag is None:
        u4, v4, ud, u1, v1 = magic_decomposition(unitary)
        ud_diag = to_bell_diagonal(ud)

    hx, hy, hz = calculate_h_vector(ud_diag)
    if np.allclose([hx, hy, hz], [0, 0, 0]):
        gatelist = [gates.Unitary(u4 @ u1, q0), gates.Unitary(v4 @ v1, q1)]
    elif np.allclose(hz, 0):
        gatelist = cnot_decomposition_light(q0, q1, hx, hy)
        if ud is None:
            return gatelist
        g0, g1 = gatelist[:2]
        gatelist[0] = gates.Unitary(g0.parameters[0] @ u1, q0)
        gatelist[1] = gates.Unitary(g1.parameters[0] @ v1, q1)

        g0, g1 = gatelist[-2:]
        gatelist[-2] = gates.Unitary(u4 @ g0.parameters[0], q0)
        gatelist[-1] = gates.Unitary(v4 @ g1.parameters[0], q1)

    else:
        cnot_dec = cnot_decomposition(q0, q1, hx, hy, hz)
        if ud is None:
            return cnot_dec

        gatelist = [
            gates.Unitary(u1, q0),
            gates.Unitary(H @ v1, q1),
        ]
        gatelist.extend(cnot_dec[1:])
        g0, g1 = gatelist[-2:]
        gatelist[-2] = gates.Unitary(u4 @ g0.parameters[0], q0)
        gatelist[-1] = gates.Unitary(v4 @ g1.parameters[0], q1)

    return gatelist



""" Qblox Cluster QRM-RF driver."""

import json

import numpy as np
from qblox_instruments.qcodes_drivers.cluster import Cluster as QbloxCluster
from qblox_instruments.qcodes_drivers.qcm_qrm import QcmQrm as QbloxQrmQcm
from qibo.config import log

from qibolab.instruments.abstract import Instrument
from qibolab.instruments.qblox.port import (
    ClusterRF_OutputPort,
    ClusterRF_OutputPort_Settings,
    QbloxInputPort,
    QbloxInputPort_Settings,
)
from qibolab.instruments.qblox.q1asm import (
    Block,
    Register,
    convert_phase,
    loop_block,
    wait_block,
)
from qibolab.instruments.qblox.sequencer import Sequencer, WaveformsBuffer
from qibolab.instruments.qblox.sweeper import QbloxSweeper, QbloxSweeperType
from qibolab.pulses import Pulse, PulseSequence, PulseShape, PulseType
from qibolab.sweeper import Parameter, Sweeper, SweeperType


class ClusterQRM_RF_Settings:
    def __init__(self, ports=None):
        if not ports:
            ports: dict = {
                "o1": ClusterRF_OutputPort_Settings(),
                "i1": QbloxInputPort_Settings(),
            }
        self.ports = ports


class ClusterQRM_RF(Instrument):
    """Qblox Cluster Qubit Readout Module RF driver.

    Qubit Readout Module RF (QRM-RF) is an instrument that integrates an arbitrary wave generator, a digitizer,
    a local oscillator and a mixer. It has one output and one input port. Each port has a path0 and path1 for the
    i(in-phase) and q(quadrature) components of the RF signal. The sampling rate of its ADC/DAC is 1 GSPS.
    https://www.qblox.com/cluster

    The class aims to simplify the configuration of the instrument, exposing only those parameters most frequently
    used and hiding other more complex settings.

    A reference to the underlying `qblox_instruments.qcodes_drivers.qcm_qrm.QRM_QCM` object is provided via the
    attribute `device`, allowing the advanced user to gain access to the features that are not exposed directly
    by the class.

    In order to accelerate the execution, the instrument settings are cached, so that the communication with the
    instrument only happens when the parameters change. This caching is done with the method
    `_set_device_parameter(target, *parameters, value)`.

    .. code-block:: text

        ports:
            o1:                                             # output port settings
                channel                     : L3-25a
                attenuation                 : 30                # (dB) 0 to 60, must be multiple of 2
                lo_enabled                  : true
                lo_frequency                : 6_000_000_000     # (Hz) from 2e9 to 18e9
                gain                        : 1                 # for path0 and path1 -1.0<=v<=1.0
            i1:                                             # input port settings
                channel                     : L2-5a
                acquisition_hold_off        : 130               # minimum 4ns
                acquisition_duration        : 1800

        classification_parameters:
            0: # qubit id
                rotation_angle              : 0                 # in degrees 0.0<=v<=360.0
                threshold                   : 0                 # in V
            1:
                rotation_angle              : 194.272
                threshold                   : 0.011197
            2:
                rotation_angle              : 104.002
                threshold                   : 0.012745

    Attributes:
        name (str): A unique name given to the instrument.
        address (str): IP_address:module_number; the IP address of the cluster and
            the module number.
        device (QbloxQrmQcm): A reference to the underlying `qblox_instruments.qcodes_drivers.qcm_qrm.QcmQrm` object.
            It can be used to access other features not directly exposed by this wrapper.
            https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/api_reference/qcm_qrm.html

        ports = A dictionary giving access to the input and output ports objects.

            - ports['o1']: Output port
            - ports['i1']: Input port

            - ports['o1'].channel (int | str): the id of the refrigerator channel the output port o1 is connected to.
            - ports['o1'].attenuation (int): (mapped to qrm.out0_att) Controls the attenuation applied to the output
              port. It must be a multiple of 2.
            - ports['o1'].lo_enabled (bool): (mapped to qrm.out0_in0_lo_en) Enables or disables the local oscillator.
            - ports['o1'].lo_frequency (int): (mapped to qrm.out0_in0_lo_freq) Sets the frequency of the local oscillator.
            - ports['o1'].gain (float): (mapped to qrm.sequencers[0].gain_awg_path0 and qrm.sequencers[0].gain_awg_path1)
              Sets the gain on both paths of the output port.
            - ports['o1'].hardware_mod_en (bool): (mapped to qrm.sequencers[0].mod_en_awg) Enables pulse modulation
              in hardware. When set to False, pulse modulation is done in software, at the host computer, and the
              modulated pulse waveform is uploaded to the instrument. When set to True, the envelope of the pulse
              is uploaded to the instrument and it is modulated in real time by the FPGA of the instrument, using
              the sequencer nco (numerically controlled oscillator).
            - ports['o1'].nco_freq (int): (mapped to qrm.sequencers[0].nco_freq). Mapped, but not configurable from
              the runcard.
            - ports['o1'].nco_phase_offs = (mapped to qrm.sequencers[0].nco_phase_offs). Mapped, but not configurable
              from the runcard.

            - ports['i1'].channel (int | str): the id of the refrigerator channel the input port o1 is connected to.
            - ports['i1'].acquisition_hold_off (int): Delay between the moment the readout pulse starts to be played and
              the start of the acquisition, in ns. It must be > 0 and multiple of 4.
            - ports['i1'].acquisition_duration (int): (mapped to qrm.sequencers[0].integration_length_acq) Duration
              of the pulse acquisition, in ns. It must be > 0 and multiple of 4.
            - ports['i1'].hardware_demod_en (bool): (mapped to qrm.sequencers[0].demod_en_acq) Enables demodulation
              and integration of the acquired pulses in hardware. When set to False, the filtration, demodulation
              and integration of the acquired pulses is done at the host computer. When set to True, the
              demodulation, integration and discretization of the pulse is done in real time at the FPGA of the
              instrument.

                - Sequencer 0 is used always for acquisitions and it is the first sequencer used to synthesise pulses.
                - Sequencer 1 to 6 are used as needed to synthesise simultaneous pulses on the same channel (required in
                  multiplexed readout) or when the memory of the default sequencers rans out.

        classification_parameters (dict): A dictionary containing the parameters needed classify the state of each qubit.
            from a single shot measurement:
        qubit_id (dict): the id of the qubit
            rotation_angle (float): 0   # in degrees 0.0<=v<=360.0. The angle of the rotation applied at the
            origin of the i q plane, that put the centroids of the state ``|0>`` and state ``|1>`` in a horizontal line.
            The rotation puts the centroid of state ``|1>`` to the right side of centroid of state ``|0>``.
        threshold (float): 0        # in V. The real component of the point along the horizontal line
            connecting both state centroids (after being rotated), that maximises the fidelity of the
            classification.

        channels (list): A list of the channels to which the instrument is connected.


    """

    DEFAULT_SEQUENCERS: dict = {"o1": 0, "i1": 0}
    SAMPLING_RATE: int = 1e9  # 1 GSPS
    FREQUENCY_LIMIT = 500e6  # 500 MHz

    property_wrapper = lambda parent, *parameter: property(
        lambda self: parent.device.get(parameter[0]),
        lambda self, x: parent._set_device_parameter(parent.device, *parameter, value=x),
    )
    property_wrapper.__doc__ = """A lambda function used to create properties that wrap around the device parameters
    and caches their value using `_set_device_parameter()`.
    """
    sequencer_property_wrapper = lambda parent, sequencer, *parameter: property(
        lambda self: parent.device.sequencers[sequencer].get(parameter[0]),
        lambda self, x: parent._set_device_parameter(parent.device.sequencers[sequencer], *parameter, value=x),
    )
    sequencer_property_wrapper.__doc__ = """A lambda function used to create properties that wrap around the device
    sequencer parameters and caches their value using `_set_device_parameter()`.
    """

    def __init__(self, name: str, address: str, settings: ClusterQRM_RF_Settings):
        """Initialises the instance.

        All class attributes are defined and initialised.
        """

        super().__init__(name, address)
        self.settings: ClusterQRM_RF_Settings = settings
        self.device: QbloxQrmQcm = None
        self.ports: dict = {}
        self.ports["o1"] = ClusterRF_OutputPort(module=self, sequencer_number=self.DEFAULT_SEQUENCERS["o1"], number=1)
        self.ports["i1"] = QbloxInputPort(
            module=self,
            output_sequencer_number=self.DEFAULT_SEQUENCERS["o1"],
            input_sequencer_number=self.DEFAULT_SEQUENCERS["i1"],
            number=1,
        )
        self.classification_parameters: dict = {}
        self.channels: list = []

        self._debug_folder: str = ""
        self._cluster: QbloxCluster = None
        self._input_ports_keys = ["i1"]
        self._output_ports_keys = ["o1"]
        self._sequencers: dict[Sequencer] = {"o1": []}
        self._port_channel_map: dict = {}
        self._channel_port_map: dict = {}
        self._device_parameters = {}
        self._device_num_output_ports = 1
        self._device_num_sequencers: int
        self._free_sequencers_numbers: list[int] = []
        self._used_sequencers_numbers: list[int] = []
        self._unused_sequencers_numbers: list[int] = []
        self._execution_time: float = 0

    def connect(self, cluster: QbloxCluster):
        """Connects to the instrument using the instrument settings in the runcard.

        Once connected, it creates port classes with properties mapped to various instrument
        parameters, and initialises the the underlying device parameters.
        """
        if not self.is_connected:
            if cluster:
                # save a reference to the underlying object
                self.device: QbloxQrmQcm = cluster.modules[int(self.address.split(":")[1]) - 1]
                # TODO: test connection with the module before continuing. Currently if there is no
                # connection the instruction self._set_device_parameter(self.device, "in0_att", value=0)
                # fails, providing a misleading error message

                # save reference to cluster
                self._cluster = cluster
                self.is_connected = True

                # once connected, initialise the parameters of the device to the default values
                self._set_device_parameter(self.device, "in0_att", value=0)
                self._set_device_parameter(
                    self.device, "out0_offset_path0", "out0_offset_path1", value=0
                )  # Default after reboot = 7.625
                self._set_device_parameter(
                    self.device, "scope_acq_avg_mode_en_path0", "scope_acq_avg_mode_en_path1", value=True
                )
                self._set_device_parameter(
                    self.device, "scope_acq_sequencer_select", value=self.DEFAULT_SEQUENCERS["i1"]
                )
                self._set_device_parameter(
                    self.device, "scope_acq_trigger_level_path0", "scope_acq_trigger_level_path1", value=0
                )
                self._set_device_parameter(
                    self.device, "scope_acq_trigger_mode_path0", "scope_acq_trigger_mode_path1", value="sequencer"
                )

                # initialise the parameters of the default sequencer to the default values,
                # the rest of the sequencers are not configured here, but will be configured
                # with the same parameters as the default in process_pulse_sequence()
                target = self.device.sequencers[self.DEFAULT_SEQUENCERS["o1"]]

                self._set_device_parameter(target, "channel_map_path0_out0_en", "channel_map_path1_out1_en", value=True)
                self._set_device_parameter(target, "cont_mode_en_awg_path0", "cont_mode_en_awg_path1", value=False)
                self._set_device_parameter(
                    target, "cont_mode_waveform_idx_awg_path0", "cont_mode_waveform_idx_awg_path1", value=0
                )
                self._set_device_parameter(target, "marker_ovr_en", value=True)  # Default after reboot = False
                self._set_device_parameter(target, "marker_ovr_value", value=15)  # Default after reboot = 0
                self._set_device_parameter(target, "mixer_corr_gain_ratio", value=1)
                self._set_device_parameter(target, "mixer_corr_phase_offset_degree", value=0)
                self._set_device_parameter(target, "offset_awg_path0", value=0)
                self._set_device_parameter(target, "offset_awg_path1", value=0)
                self._set_device_parameter(target, "sync_en", value=False)  # Default after reboot = False
                self._set_device_parameter(target, "upsample_rate_awg_path0", "upsample_rate_awg_path1", value=0)

                # on initialisation, disconnect all other sequencers from the ports
                self._device_num_sequencers = len(self.device.sequencers)
                for sequencer in range(1, self._device_num_sequencers):
                    self._set_device_parameter(
                        self.device.sequencers[sequencer],
                        "channel_map_path0_out0_en",
                        "channel_map_path1_out1_en",
                        value=False,
                    )  # Default after reboot = True

    def _set_device_parameter(self, target, *parameters, value):
        """Sets a parameter of the instrument, if it changed from the last stored in the cache.

        Args:
            target = an instance of qblox_instruments.qcodes_drivers.qcm_qrm.QcmQrm or
                                    qblox_instruments.qcodes_drivers.sequencer.Sequencer
            *parameters (list): A list of parameters to be cached and set.
            value = The value to set the paramters.
        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """
        if self.is_connected:
            key = target.name + "." + parameters[0]
            if not key in self._device_parameters:
                for parameter in parameters:
                    if not hasattr(target, parameter):
                        raise Exception(f"The instrument {self.name} does not have parameters {parameter}.")
                    target.set(parameter, value)
                self._device_parameters[key] = value
            elif self._device_parameters[key] != value:
                for parameter in parameters:
                    target.set(parameter, value)
                self._device_parameters[key] = value
        else:
            raise Exception(f"There is no connection to the instrument {self.name}.")

    def _erase_device_parameters_cache(self):
        """Erases the cache of instrument parameters."""
        self._device_parameters = {}

    def setup(self):
        """Configures the instrument with the settings of the runcard.

        A connection to the instrument needs to be established beforehand.
        Args:
            **kwargs: dict = A dictionary of settings loaded from the runcard:

                - kwargs['ports']['o1']['channel'] (int | str): the id of the refrigerator channel the output port o1 is
                  connected to.
                - kwargs['ports']['o1']['attenuation'] (int): [0 to 60 dBm, in multiples of 2] attenuation at the output.
                - kwargs['ports']['o1']['lo_enabled'] (bool): enable or disable local oscillator for up-conversion.
                - kwargs['ports']['o1']['lo_frequency'] (int): [2_000_000_000 to 18_000_000_000 Hz] local oscillator
                  frequency.
                - kwargs['ports']['o1']['gain'] (float): [0.0 - 1.0 unitless] gain applied prior to up-conversion. Qblox
                  recommends to keep `pulse_amplitude * gain` below 0.3 to ensure the mixers are working in their
                  linear regime, if necessary, lowering the attenuation applied at the output.
                - kwargs['ports']['o1']['hardware_mod_en'] (bool): enables Hardware Modulation. In this mode, pulses are
                  modulated to the intermediate frequency using the numerically controlled oscillator within the
                  fpga. It only requires the upload of the pulse envelope waveform.
                - kwargs['ports']['i1']['channel'] (int | str): the id of the refrigerator channel the input port i1 is
                  connected to.
                - kwargs['ports']['i1']['hardware_demod_en'] (bool): enables Hardware Demodulation. In this mode, the
                  sequencers of the fpga demodulate, integrate and classify the results for every shot. Once
                  integrated, the i and q values and the result of the classification requires much less memory,
                  so they can be stored for every shot in separate `bins` and retrieved later. Hardware Demodulation
                  also allows making multiple readouts on the same qubit at different points in the circuit, which is
                  not possible with Software Demodulation.
                - kwargs['acquisition_hold_off'] (int): [0 to 16834 ns, in multiples of 4] the time between the moment
                  the start of the readout pulse begins to be played, and the start of the acquisition. This is used
                  to account for the time of flight of the pulses from the output port to the input port.
                - kwargs['acquisition_duration'] (int): [0 to 8192 ns] the duration of the acquisition. It is limited by
                  the amount of memory available in the fpga to store i q samples.
                - kwargs['classification_parameters'][qubit_id][rotation_angle] (float): [0.0 to 359.999 deg] the angle
                  to rotate the results so that the projection on the real axis renders the maximum fidelity.
                - kwargs['classification_parameters'][qubit_id][threshold] (float): [-1.0 to 1.0 Volt] the voltage to be
                  used as threshold to classify the state of each shot.

        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """
        settings: ClusterQRM_RF_Settings = self.settings
        if self.is_connected:
            # Load settings
            port_settings: ClusterRF_OutputPort_Settings = settings.ports["o1"]
            self.ports["o1"].channel = port_settings.channel
            self._port_channel_map["o1"] = self.ports["o1"].channel
            self.ports["o1"].attenuation = port_settings.attenuation
            self.ports["o1"].lo_enabled = port_settings.lo_enabled
            self.ports["o1"].lo_frequency = port_settings.lo_frequency
            self.ports["o1"].gain = port_settings.gain
            self.ports["o1"].hardware_mod_en = port_settings.hardware_mod_en

            self.ports["o1"].nco_freq = 0
            self.ports["o1"].nco_phase_offs = 0

            port_settings: QbloxInputPort_Settings = settings.ports["i1"]
            self.ports["i1"].channel = port_settings.channel
            self._port_channel_map["i1"] = self.ports["i1"].channel
            self.ports["i1"].hardware_demod_en = port_settings.hardware_demod_en
            self.ports["i1"].acquisition_hold_off = port_settings.acquisition_hold_off
            self.ports["i1"].acquisition_duration = port_settings.acquisition_duration

            self._channel_port_map = {v: k for k, v in self._port_channel_map.items()}
            self.channels = list(self._channel_port_map.keys())
        else:
            raise Exception("The instrument cannot be set up, there is no connection")

    def _get_next_sequencer(self, port: str, frequency: int, qubits: dict, qubit: None):
        """Retrieves and configures the next avaliable sequencer.

        The parameters of the new sequencer are copied from those of the default sequencer, except for the intermediate
        frequency and classification parameters.

        Args:
            port (str):
            frequency (int):
            qubit (str|int):
        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """

        # select a new sequencer and configure it as required
        next_sequencer_number = self._free_sequencers_numbers.pop(0)
        if next_sequencer_number != self.DEFAULT_SEQUENCERS[port]:
            for parameter in self.device.sequencers[self.DEFAULT_SEQUENCERS[port]].parameters:
                # exclude read-only parameter `sequence` and others that have wrong default values (qblox bug)
                if not parameter in ["sequence", "thresholded_acq_marker_address", "thresholded_acq_trigger_address"]:
                    value = self.device.sequencers[self.DEFAULT_SEQUENCERS[port]].get(param_name=parameter)
                    if value:
                        target = self.device.sequencers[next_sequencer_number]
                        self._set_device_parameter(target, parameter, value=value)

        # if hardware demodulation is enabled, configure nco_frequency and classification parameters
        if self.ports["i1"].hardware_demod_en or self.ports["o1"].hardware_mod_en:
            self._set_device_parameter(
                self.device.sequencers[next_sequencer_number],
                "nco_freq",
                value=frequency,
                # It assumes all pulses in non_overlapping_pulses set have the same frequency.
                # Non-overlapping pulses of different frequencies on the same qubit channel, with hardware_demod_en
                # would lead to wrong results.
                # TODO: Throw error in that event or implement non_overlapping_same_frequency_pulses
                #       Even better, set the frequency before each pulse is played (would work with hardware modulation only)
            )
        # if self.ports["i1"].hardware_demod_en and qubit in self.classification_parameters:
        if self.ports["i1"].hardware_demod_en and not qubits[qubit].threshold is None:
            self._set_device_parameter(
                self.device.sequencers[next_sequencer_number],
                "thresholded_acq_rotation",
                # value=self.classification_parameters[qubit]["rotation_angle"],
                value=(qubits[qubit].iq_angle * 360 / (2 * np.pi)) % 360,
            )
            self._set_device_parameter(
                self.device.sequencers[next_sequencer_number],
                "thresholded_acq_threshold",
                # value=self.classification_parameters[qubit]["threshold"] * self.ports["i1"].acquisition_duration,
                value=qubits[qubit].threshold * self.ports["i1"].acquisition_duration,
            )
        # create sequencer wrapper
        sequencer = Sequencer(next_sequencer_number)
        sequencer.qubit = qubit
        return sequencer

    def get_if(self, pulse: Pulse):
        """Returns the intermediate frequency needed to synthesise a pulse based on the port lo frequency."""

        _rf = pulse.frequency
        _lo = self.ports[self._channel_port_map[pulse.channel]].lo_frequency
        _if = _rf - _lo
        if abs(_if) > self.FREQUENCY_LIMIT:
            raise Exception(
                f"""
            Pulse frequency {_rf:_} cannot be synthesised with current lo frequency {_lo:_}.
            The intermediate frequency {_if:_} would exceed the maximum frequency of {self.FREQUENCY_LIMIT:_}
            """
            )
        return _if

    def process_pulse_sequence(
        self,
        qubits: dict,
        instrument_pulses: PulseSequence,
        navgs: int,
        nshots: int,
        repetition_duration: int,
        sweepers=None,
    ):
        """Processes a sequence of pulses and sweepers, generating the waveforms and program required by
        the instrument to synthesise them.

        The output of the process is a list of sequencers used for each port, configured with the information
        required to play the sequence.
        The following features are supported:

            - multiplexed readout of up to 6 qubits
            - overlapping pulses
            - hardware modulation, demodulation, and classification
            - software modulation, with support for arbitrary pulses
            - software demodulation
            - binned acquisition
            -  real-time sweepers of

                - pulse frequency (requires hardware modulation)
                - pulse relative phase (requires hardware modulation)
                - pulse amplitude
                - pulse start
                - pulse duration
                - port gain
                - port offset

            - multiple readouts for the same qubit (sequence unrolling)
            - pulses of up to 8192 pairs of i, q samples
            - sequencer memory optimisation (waveforms cache)
            - extended waveform memory with the use of multiple sequencers
            - intrument parameters cache

        Args:
            instrument_pulses (PulseSequence): A collection of Pulse objects to be played by the instrument.
            navgs (int): The number of times the sequence of pulses should be executed averaging the results.
            nshots (int): The number of times the sequence of pulses should be executed without averaging.
            repetition_duration (int): The total duration of the pulse sequence execution plus the reset/relaxation time.
            sweepers (list(Sweeper)): A list of Sweeper objects to be implemented.
        """
        if sweepers is None:
            sweepers = []
        sequencer: Sequencer
        sweeper: Sweeper

        # calculate the number of bins
        num_bins = nshots
        for sweeper in sweepers:
            num_bins *= len(sweeper.values)

        # estimate the execution time
        self._execution_time = navgs * num_bins * ((repetition_duration + 1000 * len(sweepers)) * 1e-9)

        port = "o1"
        # initialise the list of free sequencer numbers to include the default for each port {'o1': 0}
        self._free_sequencers_numbers = [self.DEFAULT_SEQUENCERS[port]] + [1, 2, 3, 4, 5]

        # split the collection of instruments pulses by ports
        port_pulses: PulseSequence = instrument_pulses.get_channel_pulses(self._port_channel_map[port])

        # initialise the list of sequencers required by the port
        self._sequencers[port] = []

        if not port_pulses.is_empty:
            # split the collection of port pulses in non overlapping pulses
            non_overlapping_pulses: PulseSequence
            for non_overlapping_pulses in port_pulses.separate_overlapping_pulses():
                # each set of not overlapping pulses will be played by a separate sequencer
                # check sequencer availability
                if len(self._free_sequencers_numbers) == 0:
                    raise Exception(
                        f"The number of sequencers requried to play the sequence exceeds the number available {self._device_num_sequencers}."
                    )
                # get next sequencer
                sequencer = self._get_next_sequencer(
                    port=port,
                    frequency=self.get_if(non_overlapping_pulses[0]),
                    qubits=qubits,
                    qubit=non_overlapping_pulses[0].qubit,
                )
                # add the sequencer to the list of sequencers required by the port
                self._sequencers[port].append(sequencer)

                # make a temporary copy of the pulses to be processed
                pulses_to_be_processed = non_overlapping_pulses.shallow_copy()
                while not pulses_to_be_processed.is_empty:
                    pulse: Pulse = pulses_to_be_processed[0]
                    # attempt to save the waveforms to the sequencer waveforms buffer
                    try:
                        sequencer.waveforms_buffer.add_waveforms(pulse, self.ports[port].hardware_mod_en, sweepers)
                        sequencer.pulses.add(pulse)
                        pulses_to_be_processed.remove(pulse)

                    # if there is not enough memory in the current sequencer, use another one
                    except WaveformsBuffer.NotEnoughMemory:
                        if len(pulse.waveform_i) + len(pulse.waveform_q) > WaveformsBuffer.SIZE:
                            raise NotImplementedError(
                                f"Pulses with waveforms longer than the memory of a sequencer ({WaveformsBuffer.SIZE // 2}) are not supported."
                            )
                        if len(self._free_sequencers_numbers) == 0:
                            raise Exception(
                                f"The number of sequencers requried to play the sequence exceeds the number available {self._device_num_sequencers}."
                            )
                        # get next sequencer
                        sequencer = self._get_next_sequencer(
                            port=port,
                            frequency=self.get_if(non_overlapping_pulses[0]),
                            qubits=qubits,
                            qubit=non_overlapping_pulses[0].qubit,
                        )
                        # add the sequencer to the list of sequencers required by the port
                        self._sequencers[port].append(sequencer)

        # update the lists of used and unused sequencers that will be needed later on
        self._used_sequencers_numbers = []
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                self._used_sequencers_numbers.append(sequencer.number)
        self._unused_sequencers_numbers = []
        for n in range(self._device_num_sequencers):
            if not n in self._used_sequencers_numbers:
                self._unused_sequencers_numbers.append(n)

        # generate and store the Waveforms dictionary, the Acquisitions dictionary, the Weights and the Program
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                pulses = sequencer.pulses
                program = sequencer.program

                ## pre-process sweepers ##
                # TODO: move qibolab sweepers preprocessing to qblox controller

                # attach a sweeper attribute to the pulse so that it is easily accesible by the code that generates
                # the pseudo-assembly program
                pulse = None
                for pulse in pulses:
                    pulse.sweeper = None

                pulse_sweeper_parameters = [
                    Parameter.frequency,
                    Parameter.amplitude,
                    Parameter.duration,
                    Parameter.relative_phase,
                    Parameter.start,
                ]

                for sweeper in sweepers:
                    if sweeper.parameter in pulse_sweeper_parameters:
                        # check if this sequencer takes an active role in the sweep
                        if sweeper.pulses and set(sequencer.pulses) & set(sweeper.pulses):
                            # plays an active role
                            reference_value = None
                            if sweeper.parameter == Parameter.frequency:
                                if sequencer.pulses:
                                    reference_value = self.get_if(
                                        sequencer.pulses[0]
                                    )  # uses the frequency of the first pulse (assuming all same freq)
                            if sweeper.parameter == Parameter.amplitude:
                                for pulse in pulses:
                                    if pulse in sweeper.pulses:
                                        reference_value = pulse.amplitude  # uses the amplitude of the first pulse
                            if sweeper.parameter == Parameter.duration and pulse in sweeper.pulses:
                                # for duration sweepers bake waveforms
                                sweeper.qs = QbloxSweeper(
                                    program=program, type=QbloxSweeperType.duration, rel_values=pulse.idx_range
                                )
                            else:
                                # create QbloxSweepers and attach them to qibolab sweeper
                                if sweeper.type == SweeperType.OFFSET and reference_value:
                                    sweeper.qs = QbloxSweeper.from_sweeper(
                                        program=program, sweeper=sweeper, add_to=reference_value
                                    )
                                elif sweeper.type == SweeperType.FACTOR and reference_value:
                                    sweeper.qs = QbloxSweeper.from_sweeper(
                                        program=program, sweeper=sweeper, multiply_to=reference_value
                                    )
                                else:
                                    sweeper.qs = QbloxSweeper.from_sweeper(program=program, sweeper=sweeper)

                            # finally attach QbloxSweepers to the pulses being swept
                            sweeper.qs.update_parameters = True
                            pulse.sweeper = sweeper.qs
                        else:
                            # does not play an active role
                            sweeper.qs = QbloxSweeper(
                                program=program,
                                type=QbloxSweeperType.number,
                                rel_values=range(len(sweeper.values)),
                                name=sweeper.parameter.name,
                            )

                    # else: # qubit_sweeper_parameters
                    #     if sweeper.qubits and sequencer.qubit in [_.name for _ in sweeper.qubits]:
                    #         # plays an active role
                    #         if sweeper.parameter == Parameter.bias:
                    #             reference_value = self.ports[port].offset
                    #             # create QbloxSweepers and attach them to qibolab sweeper
                    #             if sweeper.type == SweeperType.ABSOLUTE:
                    #                 sweeper.qs = QbloxSweeper.from_sweeper(
                    #                     program=program, sweeper=sweeper, add_to=-reference_value
                    #                 )
                    #             elif sweeper.type == SweeperType.OFFSET:
                    #                 sweeper.qs = QbloxSweeper.from_sweeper(program=program, sweeper=sweeper)
                    #             elif sweeper.type == SweeperType.FACTOR:
                    #                 raise Exception("SweeperType.FACTOR for Parameter.bias not supported")
                    #             sweeper.qs.update_parameters = True
                    #     else:
                    #         # does not play an active role
                    #         sweeper.qs = QbloxSweeper(
                    #             program=program, type=QbloxSweeperType.number, rel_values=range(len(sweeper.values)),
                    #             name = sweeper.parameter.name
                    #         )
                    else:
                        # does not play an active role
                        sweeper.qs = QbloxSweeper(
                            program=program,
                            type=QbloxSweeperType.number,
                            rel_values=range(len(sweeper.values)),
                            name=sweeper.parameter.name,
                        )

                    # # FIXME: for qubit sweepers (Parameter.bias, Parameter.attenuation, Parameter.gain), the qubit
                    # # information alone is not enough to determine what instrument parameter is to be swept.
                    # # For example port gain, both the drive and readout ports have gain parameters.
                    # # Until this is resolved, and since bias is only implemented with QCMs offset, this instrument will
                    # # never take an active role in those sweeps.

                # Waveforms
                for index, waveform in enumerate(sequencer.waveforms_buffer.unique_waveforms):
                    sequencer.waveforms[waveform.serial] = {"data": waveform.data.tolist(), "index": index}

                # Acquisitions
                for acquisition_index, pulse in enumerate(sequencer.pulses.ro_pulses):
                    sequencer.acquisitions[pulse.serial] = {"num_bins": num_bins, "index": acquisition_index}

                # Add scope_acquisition to default sequencer
                if sequencer.number == self.DEFAULT_SEQUENCERS[port]:
                    sequencer.acquisitions["scope_acquisition"] = {"num_bins": 1, "index": acquisition_index + 1}

                # Program
                minimum_delay_between_instructions = 4

                # Active reset is not fully tested yet
                active_reset = False
                active_reset_address = 1
                active_reset_pulse_idx_I = 1
                active_reset_pulse_idx_Q = 1

                sequence_total_duration = pulses.finish  # the minimum delay between instructions is 4ns
                time_between_repetitions = repetition_duration - sequence_total_duration
                assert time_between_repetitions > minimum_delay_between_instructions
                # TODO: currently relaxation_time needs to be greater than acquisition_hold_off
                # so that the time_between_repetitions is equal to the sequence_total_duration + relaxation_time
                # to be compatible with th erest of the platforms, change it so that time_between_repetitions
                # is equal to pulsesequence duration + acquisition_hold_off if relaxation_time < acquisition_hold_off

                # create registers for key variables
                # nshots is used in the loop that iterates over the number of shots
                nshots_register = Register(program, "nshots")
                # during a sweep, each shot is saved in the bin bin_n
                bin_n = Register(program, "bin_n")
                # navgs is used in the loop of hardware averages
                navgs_register = Register(program, "navgs")

                header_block = Block("setup")
                if active_reset:
                    header_block.append(
                        f"set_latch_en {active_reset_address}, 4", f"monitor triggers on address {active_reset_address}"
                    )

                body_block = Block()

                body_block.append(f"wait_sync {minimum_delay_between_instructions}")
                if self.ports["i1"].hardware_demod_en or self.ports["o1"].hardware_mod_en:
                    body_block.append("reset_ph")
                    body_block.append_spacer()

                pulses_block = Block("play_and_acquire")
                # Add an initial wait instruction for the first pulse of the sequence
                initial_wait_block = wait_block(
                    wait_time=pulses[0].start, register=Register(program), force_multiples_of_four=True
                )
                pulses_block += initial_wait_block

                for n in range(pulses.count):
                    if pulses[n].sweeper and pulses[n].sweeper.type == QbloxSweeperType.start:
                        pulses_block.append(f"wait {pulses[n].sweeper.register}")

                    if self.ports["o1"].hardware_mod_en:
                        # # Set frequency
                        # _if = self.get_if(pulses[n])
                        # pulses_block.append(f"set_freq {convert_frequency(_if)}", f"set intermediate frequency to {_if} Hz")

                        # Set phase
                        if pulses[n].sweeper and pulses[n].sweeper.type == QbloxSweeperType.relative_phase:
                            pulses_block.append(f"set_ph {pulses[n].sweeper.register}")
                        else:
                            pulses_block.append(
                                f"set_ph {convert_phase(pulses[n].relative_phase)}",
                                comment=f"set relative phase {pulses[n].relative_phase} rads",
                            )

                    if pulses[n].type == PulseType.READOUT:
                        delay_after_play = self.ports["i1"].acquisition_hold_off

                        if len(pulses) > n + 1:
                            # If there are more pulses to be played, the delay is the time between the pulse end and the next pulse start
                            delay_after_acquire = (
                                pulses[n + 1].start - pulses[n].start - self.ports["i1"].acquisition_hold_off
                            )
                        else:
                            delay_after_acquire = sequence_total_duration - pulses[n].start
                            time_between_repetitions = (
                                repetition_duration - sequence_total_duration - self.ports["i1"].acquisition_hold_off
                            )
                            assert time_between_repetitions > 0

                        if delay_after_acquire < minimum_delay_between_instructions:
                            raise Exception(
                                f"The minimum delay after starting acquisition is {minimum_delay_between_instructions}ns."
                            )

                        if pulses[n].sweeper and pulses[n].sweeper.type == QbloxSweeperType.duration:
                            RI = pulses[n].sweeper.register
                            if pulses[n].type == PulseType.FLUX:
                                RQ = pulses[n].sweeper.register
                            else:
                                RQ = pulses[n].sweeper.aux_register

                            pulses_block.append(
                                f"play  {RI},{RQ},{delay_after_play}",  # FIXME delay_after_play won't work as the duration increases
                                comment=f"play pulse {pulses[n]} sweeping its duration",
                            )
                        else:
                            # Prepare play instruction: play wave_i_index, wave_q_index, delay_next_instruction
                            pulses_block.append(
                                f"play  {sequencer.waveforms_buffer.unique_waveforms.index(pulses[n].waveform_i)},{sequencer.waveforms_buffer.unique_waveforms.index(pulses[n].waveform_q)},{delay_after_play}",
                                comment=f"play waveforms {pulses[n]}",
                            )

                        # Prepare acquire instruction: acquire acquisition_index, bin_index, delay_next_instruction
                        if active_reset:
                            pulses_block.append(f"acquire {pulses.ro_pulses.index(pulses[n])},{bin_n},4")
                            pulses_block.append(f"latch_rst {delay_after_acquire + 300 - 4}")
                        else:
                            pulses_block.append(
                                f"acquire {pulses.ro_pulses.index(pulses[n])},{bin_n},{delay_after_acquire}"
                            )

                    else:
                        # Calculate the delay_after_play that is to be used as an argument to the play instruction
                        if len(pulses) > n + 1:
                            # If there are more pulses to be played, the delay is the time between the pulse end and the next pulse start
                            delay_after_play = pulses[n + 1].start - pulses[n].start
                        else:
                            delay_after_play = sequence_total_duration - pulses[n].start

                        if delay_after_play < minimum_delay_between_instructions:
                            raise Exception(
                                f"The minimum delay between the start of two pulses in the same channel is {minimum_delay_between_instructions}ns."
                            )

                        if pulses[n].sweeper and pulses[n].sweeper.type == QbloxSweeperType.duration:
                            RI = pulses[n].sweeper.register
                            if pulses[n].type == PulseType.FLUX:
                                RQ = pulses[n].sweeper.register
                            else:
                                RQ = pulses[n].sweeper.aux_register

                            pulses_block.append(
                                f"play  {RI},{RQ},{delay_after_play}",  # FIXME delay_after_play won't work as the duration increases
                                comment=f"play pulse {pulses[n]} sweeping its duration",
                            )
                        else:
                            # Prepare play instruction: play wave_i_index, wave_q_index, delay_next_instruction
                            pulses_block.append(
                                f"play  {sequencer.waveforms_buffer.unique_waveforms.index(pulses[n].waveform_i)},{sequencer.waveforms_buffer.unique_waveforms.index(pulses[n].waveform_q)},{delay_after_play}",
                                comment=f"play waveforms {pulses[n]}",
                            )

                body_block += pulses_block
                body_block.append_spacer()

                if active_reset:
                    final_reset_block = Block()
                    final_reset_block.append(f"set_cond 1, {active_reset_address}, 0, 4", comment="active reset")
                    final_reset_block.append(f"play {active_reset_pulse_idx_I}, {active_reset_pulse_idx_Q}, 4", level=1)
                    final_reset_block.append(f"set_cond 0, {active_reset_address}, 0, 4")
                else:
                    final_reset_block = wait_block(
                        wait_time=time_between_repetitions, register=Register(program), force_multiples_of_four=False
                    )
                final_reset_block.append_spacer()
                final_reset_block.append(f"add {bin_n}, 1, {bin_n}", "increase bin counter")

                body_block += final_reset_block

                footer_block = Block("cleaup")
                footer_block.append(f"stop")

                # wrap pulses block in sweepers loop blocks
                for sweeper in sweepers:
                    body_block = sweeper.qs.block(inner_block=body_block)

                nshots_block: Block = loop_block(
                    start=0, stop=nshots, step=1, register=nshots_register, block=body_block
                )
                nshots_block.prepend(f"move 0, {bin_n}", "reset bin counter")
                nshots_block.append_spacer()

                navgs_block = loop_block(start=0, stop=navgs, step=1, register=navgs_register, block=nshots_block)
                program.add_blocks(header_block, navgs_block, footer_block)

                sequencer.program = repr(program)

    def upload(self):
        """Uploads waveforms and programs of all sequencers and arms them in preparation for execution.

        This method should be called after `process_pulse_sequence()`.
        It configures certain parameters of the instrument based on the needs of resources determined
        while processing the pulse sequence.
        """
        # Setup
        for sequencer_number in self._used_sequencers_numbers:
            target = self.device.sequencers[sequencer_number]
            self._set_device_parameter(target, "sync_en", value=True)
            self._set_device_parameter(target, "marker_ovr_en", value=True)  # Default after reboot = False
            self._set_device_parameter(target, "marker_ovr_value", value=15)  # Default after reboot = 0

        for sequencer_number in self._unused_sequencers_numbers:
            target = self.device.sequencers[sequencer_number]
            self._set_device_parameter(target, "sync_en", value=False)
            self._set_device_parameter(target, "marker_ovr_en", value=False)  # Default after reboot = False
            self._set_device_parameter(target, "marker_ovr_value", value=0)  # Default after reboot = 0
            if sequencer_number >= 1:  # Never disconnect default sequencers
                self._set_device_parameter(target, "channel_map_path0_out0_en", value=False)
                self._set_device_parameter(target, "channel_map_path1_out1_en", value=False)

        # Upload waveforms and program
        qblox_dict = {}
        sequencer: Sequencer
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                # Add sequence program and waveforms to single dictionary
                qblox_dict[sequencer] = {
                    "waveforms": sequencer.waveforms,
                    "weights": sequencer.weights,
                    "acquisitions": sequencer.acquisitions,
                    "program": sequencer.program,
                }

                # Upload dictionary to the device sequencers
                self.device.sequencers[sequencer.number].sequence(qblox_dict[sequencer])

                # DEBUG: QRM RF Save sequence to file
                if self._debug_folder != "":
                    filename = self._debug_folder + f"Z_{self.name}_sequencer{sequencer.number}_sequence.json"
                    with open(filename, "w", encoding="utf-8") as file:
                        json.dump(qblox_dict[sequencer], file, indent=4)
                        file.write(sequencer.program)

        # Clear acquisition memory and arm sequencers
        for sequencer_number in self._used_sequencers_numbers:
            self.device.sequencers[sequencer_number].delete_acquisition_data(all=True)
            self.device.arm_sequencer(sequencer_number)

        # DEBUG: QRM RF Print Readable Snapshot
        # print(self.name)
        # self.device.print_readable_snapshot(update=True)

        # DEBUG: QRM RF Save Readable Snapshot
        from qibolab.instruments.qblox.debug import print_readable_snapshot

        if self._debug_folder != "":
            filename = self._debug_folder + f"Z_{self.name}_snapshot.json"
            with open(filename, "w", encoding="utf-8") as file:
                print_readable_snapshot(self.device, file, update=True)

    def play_sequence(self):
        """Plays the sequence of pulses.

        Starts the sequencers needed to play the sequence of pulses."""

        # Start used sequencers
        for sequencer_number in self._used_sequencers_numbers:
            self.device.start_sequencer(sequencer_number)

    def acquire(self):
        """Retrieves the readout results.

        Returns:
            The results returned vary depending on whether demodulation is performed in software or hardware:
            - Software Demodulation:
              Every readout pulse triggers an acquisition, where the 16384 i and q samples of the waveform
              acquired by the ADC are saved into a dedicated memory within the FPGA. This is what qblox calls
              *scoped acquisition*. The results of multiple shots are averaged in this memory, and cannot be
              retrieved independently. The resulting waveforms averages (i and q) are then demodulated and
              integrated in software (and finally divided by the number of samples).
              Since Software Demodulation relies on the data of the scoped acquisition and that data is the
              average of all acquisitions, **only one readout pulse per qubit is supported**, so that
              the averages all correspond to reading the same quantum state.
              Multiple symultaneous readout pulses on different qubits are supported.
              The results returned are:

                - acquisition_results["averaged_raw"] (dict): a dictionary containing tuples with the averages of
                  the i and q waveforms for every readout pulse:
                  ([i samples], [q samples])
                  The data for a specific reaout pulse can be obtained either with:

                    - `acquisition_results["averaged_raw"][ro_pulse.serial]`
                    - `acquisition_results["averaged_raw"][ro_pulse.qubit]`

                - acquisition_results["averaged_demodulated_integrated"] (dict): a dictionary containing tuples
                  with the results of demodulating and integrating (averaging over time) the average of the
                  waveforms for every pulse: ``(amplitude[V], phase[rad], i[V], q[V])``

              The data for a specific readout pulse can be obtained either with:

                - `acquisition_results["averaged_demodulated_integrated"][ro_pulse.serial]`
                - `acquisition_results["averaged_demodulated_integrated"][ro_pulse.qubit]`

              Or directly with:

                - `acquisition_results[ro_pulse.serial]`
                - `acquisition_results[ro_pulse.qubit]`

            - Hardware Demodulation:
              With hardware demodulation activated, the FPGA can demodulate, integrate (average over time), and classify
              each shot individually, saving the results on separate bins. The raw data of each acquisition continues to
              be averaged as with software modulation, so there is no way to access the raw data of each shot (unless
              executed one shot at a time). The FPGA uses fixed point arithmetic for the demodulation and integration;
              if the power level of the signal at the input port is low (the minimum resolution of the ADC is 240uV)
              rounding precission errors can accumulate and render wrong results. It is advisable to have a power level
              at least higher than 5mV.
              The results returned are:

                - acquisition_results["demodulated_integrated_averaged"] (dict): a dictionary containing tuples
                  with the results of demodulating and integrating (averaging over time) each shot waveform and then
                  averaging of the many shots: ``(amplitude[V], phase[rad], i[V], q[V])``
                - acquisition_results["demodulated_integrated_binned"] (dict): a dictionary containing tuples of lists
                  with the results of demodulating and integrating every shot waveform: ``([amplitudes[V]], [phases[rad]], [is[V]], [qs[V]])``
                - acquisition_results["demodulated_integrated_classified_binned"] (dict): a dictionary containing lists
                  with the results of demodulating, integrating and classifying every shot: ``([states[0 or 1]])``
                - acquisition_results["probability"] (dict): a dictionary containing the frequency of state 1 measurements:
                      total number of shots classified as 1 / number of shots

              If the number of readout pulses per qubit is only one, then the following is also provided:

                - acquisition_results["averaged_raw"] (dict): a dictionary containing tuples with the averages of
                  the i and q waveforms for every readout pulse: ``([i samples], [q samples])``
                - acquisition_results["averaged_demodulated_integrated"] (dict): a dictionary containing tuples
                  with the results of demodulating and integrating (averaging over time) the average of the
                  waveforms for every pulse: ``(amplitude[V], phase[rad], i[V], q[V])``

              The data within each of the above dictionaries, for a specific readout pulse or for the last readout
              pulse of a qubit can be retrieved either with:

                - `acquisition_results[dictionary_name][ro_pulse.serial]`
                - `acquisition_results[dictionary_name][ro_pulse.qubit]`

              And acquisition_results["averaged_demodulated_integrated"] directly with:

                - `acquisition_results[ro_pulse.serial]`
                - `acquisition_results[ro_pulse.qubit]`

        """

        # wait until all sequencers stop
        time_out = int(self._execution_time) + 60
        import time

        from qibo.config import log

        t = time.time()

        for sequencer_number in self._used_sequencers_numbers:
            while True:
                try:
                    state = self.device.get_sequencer_state(sequencer_number)
                except:
                    pass
                else:
                    if state.status == "STOPPED":
                        # log.info(f"{self.device.sequencers[sequencer_number].name} state: {state}")
                        # TODO: check flags for errors
                        break
                    elif time.time() - t > time_out:
                        log.info(f"Timeout - {self.device.sequencers[sequencer_number].name} state: {state}")
                        self.device.stop_sequencer(sequencer_number)
                        break
                time.sleep(1)

        acquisition_results = {}

        # Qblox qrm modules only have one memory for scope acquisition.
        # Only one sequencer can save data to that memory.
        # Several acquisitions at different points in the circuit will result in the undesired averaging
        # of the results.
        # Scope Acquisition should only be used with one acquisition per module.
        # Several readout pulses are supported for as long as they take place symultaneously.
        # Scope Acquisition data should be ignored with more than one acquisition or with Hardware Demodulation.

        # Software Demodulation requires the data from Scope Acquisition, therefore Software Demodulation only works
        # with one acquisition per module.

        # The data is retrieved by storing it first in one of the acquisitions of one of the sequencers.
        # Any could be used, but we always use 'scope_acquisition' acquisition of the default sequencer to store it.

        # Store scope acquisition data on 'scope_acquisition' acquisition of the default sequencer
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                if sequencer.number == self.DEFAULT_SEQUENCERS[port]:
                    self.device.store_scope_acquisition(sequencer.number, "scope_acquisition")
                    scope_acquisition_raw_results = self.device.get_acquisitions(sequencer.number)["scope_acquisition"]

        acquisition_results["demodulated_integrated_averaged"] = {}
        acquisition_results["averaged_raw"] = {}
        acquisition_results["averaged_demodulated_integrated"] = {}
        acquisition_results["demodulated_integrated_binned"] = {}
        acquisition_results["demodulated_integrated_classified_binned"] = {}
        acquisition_results["probability"] = {}
        data = {}
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                if not self.ports["i1"].hardware_demod_en:  # Software Demodulation
                    if len(sequencer.pulses.ro_pulses) == 1:
                        pulse = sequencer.pulses.ro_pulses[0]

                        acquisition_results["averaged_raw"][pulse.serial] = (
                            scope_acquisition_raw_results["acquisition"]["scope"]["path0"]["data"][
                                0 : self.ports["i1"].acquisition_duration
                            ],
                            scope_acquisition_raw_results["acquisition"]["scope"]["path1"]["data"][
                                0 : self.ports["i1"].acquisition_duration
                            ],
                        )
                        acquisition_results["averaged_raw"][pulse.qubit] = acquisition_results["averaged_raw"][
                            pulse.serial
                        ]

                        i, q = self._process_acquisition_results(scope_acquisition_raw_results, pulse, demodulate=True)
                        acquisition_results["averaged_demodulated_integrated"][pulse.serial] = (
                            np.sqrt(i**2 + q**2),
                            np.arctan2(q, i),
                            i,
                            q,
                        )
                        acquisition_results["averaged_demodulated_integrated"][pulse.qubit] = acquisition_results[
                            "averaged_demodulated_integrated"
                        ][pulse.serial]

                        # Default Results = Averaged Demodulated Integrated
                        acquisition_results[pulse.serial] = acquisition_results["averaged_demodulated_integrated"][
                            pulse.serial
                        ]
                        acquisition_results[pulse.qubit] = acquisition_results["averaged_demodulated_integrated"][
                            pulse.qubit
                        ]

                        # ignores the data available in acquisition results and returns only i and q voltages
                        # TODO: to be updated once the functionality of ExecutionResults is extended
                        data[pulse.serial] = (i, q)

                    else:
                        raise Exception(
                            """Software Demodulation only supports one acquisition per channel.
                        Multiple readout pulses are supported as long as they are symultaneous (requiring one acquisition)"""
                        )

                else:  # Hardware Demodulation
                    binned_raw_results = self.device.get_acquisitions(sequencer.number)
                    for pulse in sequencer.pulses.ro_pulses:
                        acquisition_name = pulse.serial
                        i, q = self._process_acquisition_results(
                            binned_raw_results[acquisition_name], pulse, demodulate=False
                        )
                        acquisition_results["demodulated_integrated_averaged"][pulse.serial] = (
                            np.sqrt(i**2 + q**2),
                            np.arctan2(q, i),
                            i,
                            q,
                        )
                        acquisition_results["demodulated_integrated_averaged"][pulse.qubit] = acquisition_results[
                            "demodulated_integrated_averaged"
                        ][pulse.serial]

                        # Default Results = Demodulated Integrated Averaged
                        acquisition_results[pulse.serial] = acquisition_results["demodulated_integrated_averaged"][
                            pulse.serial
                        ]
                        acquisition_results[pulse.qubit] = acquisition_results["demodulated_integrated_averaged"][
                            pulse.qubit
                        ]

                        # Save individual shots
                        shots_i = (
                            np.array(
                                binned_raw_results[acquisition_name]["acquisition"]["bins"]["integration"]["path0"]
                            )
                            / self.ports["i1"].acquisition_duration
                        )
                        shots_q = (
                            np.array(
                                binned_raw_results[acquisition_name]["acquisition"]["bins"]["integration"]["path1"]
                            )
                            / self.ports["i1"].acquisition_duration
                        )

                        acquisition_results["demodulated_integrated_binned"][pulse.serial] = (
                            np.sqrt(shots_i**2 + shots_q**2),
                            np.arctan2(shots_q, shots_i),
                            shots_i,
                            shots_q,
                        )
                        acquisition_results["demodulated_integrated_binned"][pulse.qubit] = acquisition_results[
                            "demodulated_integrated_binned"
                        ][pulse.serial]

                        acquisition_results["demodulated_integrated_classified_binned"][
                            pulse.serial
                        ] = binned_raw_results[acquisition_name]["acquisition"]["bins"]["threshold"]
                        acquisition_results["demodulated_integrated_classified_binned"][
                            pulse.qubit
                        ] = acquisition_results["demodulated_integrated_classified_binned"][pulse.serial]

                        acquisition_results["probability"][pulse.serial] = np.mean(
                            acquisition_results["demodulated_integrated_classified_binned"][pulse.serial]
                        )
                        acquisition_results["probability"][pulse.qubit] = acquisition_results["probability"][
                            pulse.serial
                        ]

                        # Provide Scope Data for verification (assuming memory reseet is being done)
                        if len(sequencer.pulses.ro_pulses) == 1:
                            pulse = sequencer.pulses.ro_pulses[0]

                            acquisition_results["averaged_raw"][pulse.serial] = (
                                scope_acquisition_raw_results["acquisition"]["scope"]["path0"]["data"][
                                    0 : self.ports["i1"].acquisition_duration
                                ],
                                scope_acquisition_raw_results["acquisition"]["scope"]["path1"]["data"][
                                    0 : self.ports["i1"].acquisition_duration
                                ],
                            )
                            acquisition_results["averaged_raw"][pulse.qubit] = acquisition_results["averaged_raw"][
                                pulse.serial
                            ]

                            i, q = self._process_acquisition_results(
                                scope_acquisition_raw_results, pulse, demodulate=True
                            )

                            acquisition_results["averaged_demodulated_integrated"][pulse.serial] = (
                                np.sqrt(i**2 + q**2),
                                np.arctan2(q, i),
                                i,
                                q,
                            )
                            acquisition_results["averaged_demodulated_integrated"][pulse.qubit] = acquisition_results[
                                "averaged_demodulated_integrated"
                            ][pulse.serial]

                        # ignores the data available in acquisition results and returns only i and q voltages
                        # TODO: to be updated once the functionality of ExecutionResults is extended
                        data[acquisition_name] = (
                            shots_i,
                            shots_q,
                            np.array(acquisition_results["demodulated_integrated_classified_binned"][acquisition_name]),
                        )

                    # DEBUG: QRM RF Plot Incomming Pulses
                    # import qibolab.instruments.debug.incomming_pulse_plotting as pp
                    # pp.plot(raw_results)
                    # DEBUG: QRM RF Plot Acquisition_results
                    # from qibolab.debug.debug import plot_acquisition_results
                    # plot_acquisition_results(acquisition_results, pulse, savefig_filename="acquisition_results.png")
        return data

    def _process_acquisition_results(self, acquisition_results, readout_pulse: Pulse, demodulate=True):
        """Processes the results of the acquisition.

        If hardware demodulation is disabled, it demodulates and integrates the acquired pulse. If enabled,
        if processes the results as required by qblox (calculating the average by dividing the integrated results by
        the number of smaples acquired).
        """
        if demodulate:
            acquisition_frequency = self.get_if(readout_pulse)

            # DOWN Conversion
            n0 = 0
            n1 = self.ports["i1"].acquisition_duration
            input_vec_I = np.array(acquisition_results["acquisition"]["scope"]["path0"]["data"][n0:n1])
            input_vec_Q = np.array(acquisition_results["acquisition"]["scope"]["path1"]["data"][n0:n1])
            input_vec_I -= np.mean(input_vec_I)  # qblox does not remove the offsets in hardware
            input_vec_Q -= np.mean(input_vec_Q)

            modulated_i = input_vec_I
            modulated_q = input_vec_Q

            num_samples = modulated_i.shape[0]
            time = np.arange(num_samples) / PulseShape.SAMPLING_RATE

            cosalpha = np.cos(2 * np.pi * acquisition_frequency * time)
            sinalpha = np.sin(2 * np.pi * acquisition_frequency * time)
            demod_matrix = np.sqrt(2) * np.array([[cosalpha, sinalpha], [-sinalpha, cosalpha]])
            result = []
            for it, t, ii, qq in zip(np.arange(modulated_i.shape[0]), time, modulated_i, modulated_q):
                result.append(demod_matrix[:, :, it] @ np.array([ii, qq]))
            demodulated_signal = np.array(result)
            integrated_signal = np.mean(demodulated_signal, axis=0)

            # import matplotlib.pyplot as plt
            # plt.plot(input_vec_I[:400])
            # plt.plot(list(map(list, zip(*demodulated_signal)))[0][:400])
            # plt.show()
        else:
            i = np.mean(
                np.array(acquisition_results["acquisition"]["bins"]["integration"]["path0"])
                / self.ports["i1"].acquisition_duration
            )
            q = np.mean(
                np.array(acquisition_results["acquisition"]["bins"]["integration"]["path1"])
                / self.ports["i1"].acquisition_duration
            )
            integrated_signal = i, q
        return integrated_signal

    def start(self):
        """Empty method to comply with Instrument interface."""
        pass

    def stop(self):
        """Stops all sequencers"""
        try:
            self.device.stop_sequencer()
        except:
            pass

    def disconnect(self):
        """Empty method to comply with Instrument interface."""
        self._erase_device_parameters_cache()
        self._cluster = None
        self.is_connected = False



import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from more_itertools import pairwise
from qibo import gates
from qibo.config import log, raise_error
from qibo.models import Circuit

from qibolab.transpilers.abstract import Router, find_gates_qubits_pairs
from qibolab.transpilers.placer import assert_placement


class ConnectivityError(Exception):
    """Raise for an error in the connectivity"""


def assert_connectivity(connectivity: nx.Graph, circuit: Circuit):
    """Assert if a circuit can be executed on Hardware.
    No gates acting on more than two qubits.
    All two qubit operations can be performed on hardware

    Args:
        circuit (qibo.models.Circuit): circuit model to check.
        connectivity (networkx.graph): chip connectivity.
    """

    for gate in circuit.queue:
        if len(gate.qubits) > 2 and not isinstance(gate, gates.M):
            raise ConnectivityError(f"{gate.name} acts on more than two qubits.")
        if len(gate.qubits) == 2:
            if ("q" + str(gate.qubits[0]), "q" + str(gate.qubits[1])) not in connectivity.edges:
                raise ConnectivityError("Circuit does not respect connectivity. " f"{gate.name} acts on {gate.qubits}.")


def remap_circuit(circuit: Circuit, qubit_map):
    """Map logical to physical qubits in a circuit.

    Args:
        circuit (qibo.models.Circuit): qibo circuit to be remapped.
        qubit_map (np.array): new qubit mapping.

    Returns:
        new_circuit (qibo.models.Circuit): transpiled circuit mapped with initial qubit mapping.
    """
    new_circuit = Circuit(circuit.nqubits)
    for gate in circuit.queue:
        new_circuit.add(gate.on_qubits({q: qubit_map[q] for q in gate.qubits}))
    return new_circuit


class ShortestPaths(Router):
    """A class to perform initial qubit mapping and connectivity matching.

    Properties:
        sampling_split (float): fraction of paths tested (between 0 and 1).

    Attributes:
        connectivity (networkx.Graph): chip connectivity.
        verbose (bool): print info messages.
        initial_layout (dict): initial physical to logical qubit mapping
        added_swaps (int): number of swaps added to the circuit to match connectivity.
        _gates_qubits_pairs (list): quantum circuit represented as a list (only 2 qubit gates).
        _mapping (dict): circuit to physical qubit mapping during transpiling.
        _graph (networkx.graph): qubit mapped as nodes of the connectivity graph.
        _qubit_map (np.array): circuit to physical qubit mapping during transpiling as vector.
        _circuit_position (int): position in the circuit.

    """

    def __init__(self, connectivity: nx.Graph, sampling_split=1.0, verbose=False):
        self.connectivity = connectivity
        self.sampling_split = sampling_split
        self.verbose = verbose
        self.initial_layout = None
        self.added_swaps = 0
        self.final_map = None
        self._gates_qubits_pairs = None
        self._mapping = None
        self._graph = None
        self._qubit_map = None
        self._transpiled_circuit = None
        self._circuit_position = 0

    def __call__(self, circuit, initial_layout):
        """Circuit connectivity matching.

        Args:
            circuit (qibo.models.Circuit): circuit to be matched to hardware connectivity.
            initial_layout (dict): initial qubit mapping.

        Returns:
            hardware_mapped_circuit (qibo.models.Circuit): circut mapped to hardware topology.
            final_mapping (dict): final qubit mapping.
        """
        self._mapping = initial_layout
        init_qubit_map = np.asarray(list(initial_layout.values()))
        self.initial_checks(circuit.nqubits)
        self._gates_qubits_pairs = find_gates_qubits_pairs(circuit)
        self._graph = nx.relabel_nodes(self.connectivity, self._mapping)
        self._qubit_map = np.sort(init_qubit_map)
        self.first_transpiler_step(circuit)
        while len(self._gates_qubits_pairs) != 0:
            self.transpiler_step(circuit)
        final_mapping = {
            key: self._qubit_map[init_qubit_map[i]] for i, key in enumerate(list(self.connectivity.nodes()))
        }
        hardware_mapped_circuit = remap_circuit(self._transpiled_circuit, np.argsort(init_qubit_map))
        return hardware_mapped_circuit, final_mapping

    def transpiler_step(self, qibo_circuit):
        """Transpilation step. Find new mapping, add swap gates and apply gates that can be run with this configuration.

        Args:
            qibo_circuit (qibo.models.Circuit): circuit to be transpiled.
        """
        len_before_step = len(self._gates_qubits_pairs)
        path, meeting_point = self.relocate()
        self.add_swaps(path, meeting_point)
        self.update_qubit_map()
        self.add_gates(qibo_circuit, len_before_step - len(self._gates_qubits_pairs))

    def first_transpiler_step(self, qibo_circuit):
        """First transpilation step. Apply gates that can be run with the initial qubit mapping.

        Args:
            qibo_circuit (qibo.models.Circuit): circuit to be transpiled.
        """
        self._circuit_position = 0
        self.added_swaps = 0
        len_2q_circuit = len(self._gates_qubits_pairs)
        self._gates_qubits_pairs = self.reduce(self._graph)
        self.add_gates(qibo_circuit, len_2q_circuit - len(self._gates_qubits_pairs))

    @property
    def sampling_split(self):
        return self._sampling_split

    @sampling_split.setter
    def sampling_split(self, sampling_split):
        """Set the sampling split.

        Args:
            sampling_split (float): define fraction of shortest path tested.
        """

        if 0.0 < sampling_split <= 1.0:
            self._sampling_split = sampling_split
        else:
            raise_error(ValueError, "Sampling_split must be in (0:1].")

    def draw_connectivity(self):  # pragma: no cover
        """Draw connectivity graph."""
        position = nx.spectral_layout(self.connectivity)
        nx.draw(self.connectivity, pos=position, with_labels=True)
        plt.show()

    def reduce(self, graph):
        """Reduce the circuit, delete a 2-qubit gate if it can be applied on the current configuration.

        Args:
            graph (networkx.Graph): current hardware qubit mapping.

        Returns:
            new_circuit (list): reduced circuit.
        """
        new_circuit = self._gates_qubits_pairs.copy()
        while new_circuit != [] and (new_circuit[0][0], new_circuit[0][1]) in graph.edges():
            del new_circuit[0]
        return new_circuit

    def map_list(self, path):
        """Return all possible walks of qubits, or a fraction, for a given path.

        Args:
            path (list): path to move qubits.

        Returns:
            mapping_list (list): all possible walks of qubits, or a fraction of them based on self.sampling_split, for a given path.
            meeting_point_list (list): qubit meeting point for each path.
        """
        path_ends = [path[0], path[-1]]
        path_middle = path[1:-1]
        mapping_list = []
        meeting_point_list = []
        test_paths = range(len(path) - 1)
        if self.sampling_split != 1.0:
            test_paths = np.random.choice(
                test_paths, size=int(np.ceil(len(test_paths) * self.sampling_split)), replace=False
            )
        for i in test_paths:
            values = path_middle[:i] + path_ends + path_middle[i:]
            mapping = dict(zip(path, values))
            mapping_list.append(mapping)
            meeting_point_list.append(i)
        return mapping_list, meeting_point_list

    def relocate(self):
        """A small greedy algorithm to decide which path to take, and how qubits should walk.

        Returns:
            final_path (list): best path to move qubits.
            meeting_point (int): qubit meeting point in the path.
        """
        nodes = self._graph.number_of_nodes()
        circuit = self.reduce(self._graph)
        final_circuit = circuit
        keys = list(range(nodes))
        final_graph = self._graph
        final_mapping = dict(zip(keys, keys))
        # Consider all shortest paths
        path_list = [p for p in nx.all_shortest_paths(self._graph, source=circuit[0][0], target=circuit[0][1])]
        self.added_swaps += len(path_list[0]) - 2
        # Here test all paths
        for path in path_list:
            # map_list uses self.sampling_split
            list_, meeting_point_list = self.map_list(path)
            for j, mapping in enumerate(list_):
                new_graph = nx.relabel_nodes(self._graph, mapping)
                new_circuit = self.reduce(new_graph)
                # Greedy looking for the optimal path and the optimal walk on this path
                if len(new_circuit) < len(final_circuit):
                    final_graph = new_graph
                    final_circuit = new_circuit
                    final_mapping = mapping
                    final_path = path
                    meeting_point = meeting_point_list[j]
        self._graph = final_graph
        self._mapping = final_mapping
        self._gates_qubits_pairs = final_circuit
        return final_path, meeting_point

    def initial_checks(self, qubits):
        """Initialize the transpiled circuit and check if it can be mapped to the defined connectivity.

        Args:
            qubits (int): number of qubits in the circuit to be transpiled.
        """
        nodes = self.connectivity.number_of_nodes()
        if qubits > nodes:
            raise_error(ValueError, "There are not enough physical qubits in the hardware to map the circuit.")
        if qubits == nodes:
            new_circuit = Circuit(nodes)
        else:
            if self.verbose:
                log.info(
                    "You are using more physical qubits than required by the circuit, some ancillary qubits will be added to the circuit."
                )
            new_circuit = Circuit(nodes)
        assert_placement(new_circuit, self._mapping)
        self._transpiled_circuit = new_circuit

    def add_gates(self, qibo_circuit, matched_gates):
        """Add one and two qubit gates to transpiled circuit until connectivity is matched.

        Args:
            qibo_circuit (qibo.models.Circuit): circuit to be transpiled.
            matched_gates (int): number of two qubit gates that can be applied with the current qubit mapping.
        """
        index = 0
        while self._circuit_position < len(qibo_circuit.queue):
            gate = qibo_circuit.queue[self._circuit_position]
            if len(gate.qubits) == 1:
                self._transpiled_circuit.add(gate.on_qubits({gate.qubits[0]: self._qubit_map[gate.qubits[0]]}))
                self._circuit_position += 1
            else:
                index += 1
                if index == matched_gates + 1:
                    break
                self._transpiled_circuit.add(
                    gate.on_qubits(
                        {
                            gate.qubits[0]: self._qubit_map[gate.qubits[0]],
                            gate.qubits[1]: self._qubit_map[gate.qubits[1]],
                        }
                    )
                )
                self._circuit_position += 1

    def add_swaps(self, path, meeting_point):
        """Add swaps to the transpiled circuit to move qubits.

        Args:
            path (list): path to move qubits.
            meeting_point (int): qubit meeting point in the path.
        """
        forward = path[0 : meeting_point + 1]
        backward = list(reversed(path[meeting_point + 1 :]))
        if len(forward) > 1:
            for f1, f2 in pairwise(forward):
                self._transpiled_circuit.add(gates.SWAP(self._qubit_map[f1], self._qubit_map[f2]))
        if len(backward) > 1:
            for b1, b2 in pairwise(backward):
                self._transpiled_circuit.add(gates.SWAP(self._qubit_map[b1], self._qubit_map[b2]))

    def update_qubit_map(self):
        """Update the qubit mapping after adding swaps."""
        old_mapping = self._qubit_map.copy()
        for key, value in self._mapping.items():
            self._qubit_map[value] = old_mapping[key]


class Sabre(Router):
    # TODO: requires block circuit
    """
    Routing algorithm proposed in
    https://doi.org/10.48550/arXiv.1809.02573
    """

    def __init__(self):  # pragma: no cover
        super().__init__()



"""A library to support generating qblox q1asm programs."""

import numpy as np

END_OF_LINE = "\n"


class Program:
    """A class to represent a sequencer q1asm assembly program.

    A q1asm program is made of blocks of code (:class:`qibolab.instruments.qblox.qblox_q1asm.Block`).
    Real time registers (variables) are necessary to implement sweeps in real time.
    This class keeps track of the registers used and has a method to provide the next available register.
    """

    MAX_REGISTERS = 64

    def next_register(self):
        """Provides the number of the next available register."""
        self._next_register_number += 1
        if self._next_register_number >= Program.MAX_REGISTERS:
            raise RuntimeError("There are no more registers available.")
        return self._next_register_number

    def __init__(self):
        self._blocks: list = []
        self._next_register_number: int = -1

    def add_blocks(self, *blocks):
        """Adds a :class:`qibolab.instruments.qblox.qblox_q1asm.Block` of code to the list of blocks."""
        for block in blocks:
            self._blocks.append(block)

    def __repr__(self) -> str:
        """Returns the program."""
        block_str: str = ""
        for block in self._blocks:
            block_str += repr(block) + END_OF_LINE
        return block_str


class Block:
    """A class to represent a block of q1asm assembly code.

    A block is comprised of code lines.
    Attributes:
        name (str): A name for the block of code.
        lines (list(tupple)): A list of lines of code (code, comment, indentation level).
    """

    # TODO: replace tupples used for code lines with a Line object

    GLOBAL_INDENTATION_LEVEL = 3
    SPACES_PER_LEVEL = 4
    SPACES_BEFORE_COMMENT = 4

    def __init__(self, name=""):
        self.name = name
        self.lines: list = []
        self._indentation = 0

    def _indentation_string(self, level):
        return " " * Block.SPACES_PER_LEVEL * level

    @property
    def indentation(self):
        return self._indentation

    @indentation.setter
    def indentation(self, value):
        if not isinstance(value, int):
            raise TypeError(f"indentation type should be int, got {type(value).__name__}")

        diff = value - self._indentation
        self._indentation = value
        self.lines = [(line, comment, level + diff) for (line, comment, level) in self.lines]

    def append(self, line, comment="", level=0):
        self.lines = self.lines + [(line, comment, self._indentation + level)]

    def prepend(self, line, comment="", level=0):
        self.lines = [(line, comment, self._indentation + level)] + self.lines

    def append_spacer(self):
        self.lines = self.lines + [("", "", self._indentation)]

    def __repr__(self) -> str:
        """Returns a string with the block of code, taking care of the indentation of instructions and comments."""

        def comment_col(line, level):
            col = Block.SPACES_PER_LEVEL * (level + Block.GLOBAL_INDENTATION_LEVEL)
            col += len(line)
            return col

        max_col: int = 0
        for line, comment, level in self.lines:
            if comment:
                max_col = max(max_col, comment_col(line, level))

        block_str: str = ""
        if self.name:
            block_str += (
                self._indentation_string(self._indentation + Block.GLOBAL_INDENTATION_LEVEL)
                + "# "
                + self.name
                + END_OF_LINE
            )

        for line, comment, level in self.lines:
            block_str += self._indentation_string(level + Block.GLOBAL_INDENTATION_LEVEL) + line
            if comment:
                block_str += " " * (max_col - comment_col(line, level) + Block.SPACES_BEFORE_COMMENT) + " # " + comment
            block_str += END_OF_LINE

        return block_str

    def __add__(self, other):
        if isinstance(other, Block):
            block = Block()
            for line, comment, level in self.lines:
                block.append(line, comment, level)
            for line, comment, level in other.lines:
                block.append(line, comment, level)
        else:
            raise TypeError(f"Expected Block, got {type(other).__name__}")
        return block

    def __radd__(self, other):
        if isinstance(other, Block):
            block = Block()
            for line, comment, level in other.lines:
                block.append(line, comment, level)
            for line, comment, level in self.lines:
                block.append(line, comment, level)
        else:
            raise TypeError(f"Expected Block, got {type(other).__name__}")
        return block

    def __iadd__(self, other):
        if isinstance(other, Block):
            for line, comment, level in other.lines:
                self.append(line, comment, level)
        else:
            raise TypeError(f"Expected Block, got {type(other).__name__}")
        return self


class Register:
    """A class to represent a q1asm program register.

    Registers are used as variables in real time FPGA code.
    Attributes:
        name (str): a name to identify the register
    """

    def __init__(self, program: Program, name: str = ""):
        self._number = program.next_register()
        self._name = name
        self._type = type

    def __repr__(self) -> str:
        return "R" + str(self._number)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value


def wait_block(wait_time: int, register: Register, force_multiples_of_four: bool = False):
    """Generates blocks of code to implement long delays.

    Arguments:
        wait_time (int): the total time to wait.
        register (:class:`qibolab.instruments.qblox.qblox_q1asm.Register`): the register used to loop
        force_multiples_of_four (bool): a flag that forces the delay to be a multiple of 4(ns)
    """
    n_loops: int
    loop_wait: int
    wait: int
    block = Block()

    # constrains
    # extra_wait and wait_loop_step need to be within (4,65535) (2**16 bits variable)
    # extra_wait and wait_loop_step need to be multiples of 4ns *

    if wait_time < 0:
        raise ValueError("wait_time must be positive.")

    elif wait_time == 0:
        n_loops = 0
        loop_wait = 0
        wait = 0

    elif wait_time > 0 and wait_time < 4:
        # TODO: log("wait_time > 0 and wait_time < 4 is not supported by the instrument, wait_time changed to 4ns")
        n_loops = 0
        loop_wait = 0
        wait = 4

    elif wait_time >= 4 and wait_time < 2**16:  # 65536ns
        n_loops = 0
        loop_wait = 0
        wait = wait_time
    elif wait_time >= 2**16 and wait_time < 2**32:  # 4.29s
        loop_wait = 2**16 - 4
        n_loops = wait_time // loop_wait
        wait = wait_time % loop_wait

        while loop_wait >= 4 and (wait > 0 and wait < 4):
            loop_wait -= 4
            n_loops = wait_time // loop_wait
            wait = wait_time % loop_wait

        if loop_wait < 4 or (wait > 0 and wait < 4):
            raise ValueError(f"Unable to decompose {wait_time} into valid (n_loops * loop_wait + wait)")
    else:
        raise ValueError("wait_time > 65535**2 is not supported yet.")

    if force_multiples_of_four:
        wait = int(np.ceil(wait / 4)) * 4

    wait_time = n_loops * loop_wait + wait

    if n_loops > 0:
        block.append(f"# wait {wait_time} ns")
        block.append(f"move {n_loops}, {register}")
        block.append(f"wait_loop_{register}:")
        block.append(f"wait {loop_wait}", level=1)
        block.append(f"loop {register}, @wait_loop_{register}", level=1)
    if wait > 0:
        block.append(f"wait {wait}")

    return block


def loop_block(start: int, stop: int, step: int, register: Register, block: Block):  # validate values
    """Generates blocks of code to implement loops.

    Its behaviour is similar to range(): it includes the first value, but never the last.

    Arguments:
        start (int): the initial value.
        stop (int): the final value (excluded).
        step (int): the step in which the register is incremented.
        register (:class:`qibolab.instruments.qblox.qblox_q1asm.Register`): the register modified within the loop.
        block (:class:`qibolab.instruments.qblox.qblox_q1asm.Register`): the block of code to be iterated.
    """

    if not (isinstance(start, int) and isinstance(stop, int) and isinstance(step, int)):
        raise ValueError("begin, end and step must be int")
    if stop != start and step == 0:
        raise ValueError("step must not be 0")
    if (stop > start and not step > 0) or (stop < start and not step < 0):
        raise ValueError("invalid step")
    if start < 0 or stop < 0:
        raise ValueError("sweep possitive values only")

    header_block = Block()
    # same behaviour as range() includes the first but never the last
    header_block.append(f"move {start}, {register}", comment=register.name + " loop")
    header_block.append("nop")
    header_block.append(f"loop_{register}:")
    header_block.append_spacer()

    body_block = Block()
    body_block.indentation = 1
    body_block += block

    footer_block = Block()
    footer_block.append_spacer()
    if stop > start:
        footer_block.append(f"add {register}, {step}, {register}")
        footer_block.append("nop")
        footer_block.append(f"jlt {register}, {stop}, @loop_{register}", comment=register.name + " loop")
    elif stop < start:
        footer_block.append(f"sub {register}, {abs(step)}, {register}")
        footer_block.append("nop")
        footer_block.append(f"jge {register}, {stop + 1}, @loop_{register}", comment=register.name + " loop")

    return header_block + body_block + footer_block


def convert_phase(phase_rad: float):
    """Converts phase values in radiants to the encoding used in qblox FPGAs.

    The phase is divided into 1e9 steps between 0° and 360°,
    expressed as an integer between 0 and 1e9 (e.g 45°=125e6).
    https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/api_reference/sequencer.html
    """
    phase_deg = (phase_rad * 360 / (2 * np.pi)) % 360
    return int(phase_deg * 1e9 / 360)


def convert_frequency(freq: float):
    """Converts frequency values to the encoding used in qblox FPGAs.

    The frequency is divided into 4e9 steps between -500 and 500 MHz and
    expressed as an integer between -2e9 and 2e9. (e.g. 1 MHz=4e6).
    https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/api_reference/sequencer.html
    """
    if not (freq >= -500e6 and freq <= 500e6):
        raise ValueError("frequency must be a float between -500e6 and 500e6 Hz")
    return int(freq * 4) % 2**32  # two's complement of 18? TODO: confirm with qblox


def convert_gain(gain: float):
    """Converts gain values to the encoding used in qblox FPGAs.

    Both gain values are divided in 2**sample path width steps.
    QCM DACs resolution 16bits, QRM DACs and ADCs 12 bit
    https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/api_reference/sequencer.html
    """
    if not (gain >= -1 and gain <= 1):
        raise ValueError("gain must be a float between -1 and 1")
    if gain == 1:
        return 2**15 - 1
    else:
        return int(np.floor(gain * 2**15)) % 2**32  # two's complement 32 bit number


def convert_offset(offset: float):
    """Converts offset values to the encoding used in qblox FPGAs.

    Both offset values are divided in 2**sample path width steps.
    QCM DACs resolution 16bits, QRM DACs and ADCs 12 bit
    QCM 5Vpp, QRM 2Vpp
    https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/api_reference/sequencer.html
    """
    scale_factor = 1.25 * np.sqrt(2)
    normalised_offset = offset / scale_factor

    if not (normalised_offset >= -1 and normalised_offset <= 1):
        raise ValueError(f"offset must be a float between {-scale_factor:.3f} and {scale_factor:.3f} V")
    if normalised_offset == 1:
        return 2**15 - 1
    else:
        return int(np.floor(normalised_offset * 2**15)) % 2**32  # two's complement 32 bit number? or 12 or 24?



"""A platform for executing quantum algorithms."""

import math
import re
from dataclasses import dataclass, field, replace
from typing import Dict, List, Optional

import networkx as nx
from qibo.config import log, raise_error

from qibolab.execution_parameters import ExecutionParameters
from qibolab.instruments.abstract import Controller, Instrument, InstrumentId
from qibolab.native import NativeType
from qibolab.pulses import PulseSequence
from qibolab.qubits import Qubit, QubitId, QubitPair, QubitPairId
from qibolab.sweeper import Sweeper

InstrumentMap = Dict[InstrumentId, Instrument]
QubitMap = Dict[QubitId, Qubit]
QubitPairMap = Dict[QubitPairId, QubitPair]


@dataclass
class Settings:
    """Default execution settings read from the runcard."""

    nshots: int = 1024
    """Default number of repetitions when executing a pulse sequence."""
    sampling_rate: int = int(1e9)
    """Number of waveform samples supported by the instruments per second."""
    relaxation_time: int = int(1e5)
    """Time in ns to wait for the qubit to relax to its ground state between shots."""


@dataclass
class Platform:
    """Platform for controlling quantum devices."""

    name: str
    """Name of the platform."""
    qubits: QubitMap
    """Dictionary mapping qubit names to :class:`qibolab.qubits.Qubit` objects."""
    pairs: QubitPairMap
    """Dictionary mapping sorted tuples of qubit names to :class:`qibolab.qubits.QubitPair` objects."""
    instruments: InstrumentMap
    """Dictionary mapping instrument names to :class:`qibolab.instruments.abstract.Instrument` objects."""

    settings: Settings = field(default_factory=Settings)
    """Container with default execution settings."""
    resonator_type: Optional[str] = None
    """Type of resonator (2D or 3D) in the used QPU.
    Default is 3D for single-qubit chips and 2D for multi-qubit.
    """

    is_connected: bool = False
    """Flag for whether we are connected to the physical instruments."""
    two_qubit_native_types: NativeType = field(default_factory=lambda: NativeType(0))
    """Types of two qubit native gates. Used by the transpiler."""
    topology: nx.Graph = field(default_factory=nx.Graph)
    """Graph representing the qubit connectivity in the quantum chip."""

    def __post_init__(self):
        log.info("Loading platform %s", self.name)
        if self.resonator_type is None:
            self.resonator_type = "3D" if self.nqubits == 1 else "2D"

        for pair in self.pairs.values():
            self.two_qubit_native_types |= pair.native_gates.types
        if self.two_qubit_native_types is NativeType(0):
            # dummy value to avoid transpiler failure for single qubit devices
            self.two_qubit_native_types = NativeType.CZ

        self.topology.add_nodes_from(self.qubits.keys())
        self.topology.add_edges_from([(pair.qubit1.name, pair.qubit2.name) for pair in self.pairs.values()])

    def __str__(self):
        return self.name

    @property
    def nqubits(self) -> int:
        """Total number of usable qubits in the QPU.."""
        # TODO: Seperate couplers from qubits (PR #508)
        return len([qubit for qubit in self.qubits if not (isinstance(qubit, str) and "c" in qubit)])

    def update(self, updates: dict):
        r"""Updates platform common runcard parameters after calibration actions.

        Args:
            updates (dict): Dictionary containing the parameters to update the runcard.
                            A typical dictionary should be of the following form
                            {`parameter_to_update_in_runcard`: {`qubit0`:`par_value_qubit0`, ..., `qubit_i`:`par_value_qubit_i`, ...}}.
                            The parameters that can be updated by this method are:

                            - readout_frequency (GHz)
                            - readout_attenuation (dimensionless)
                            - bare_resonator_frequency (GHz)
                            - sweetspot(V)
                            - drive_frequency (GHz)
                            - readout_amplitude (dimensionless)
                            - drive_amplitude (dimensionless)
                            - drive_length
                            - t2 (ns)
                            - t2_spin_echo (ns)
                            - t1 (ns)
                            - thresold(V)
                            - iq_angle(deg)
                            - mean_gnd_states(V)
                            - mean_exc_states(V)
                            - beta(dimensionless)

        """
        for par, values in updates.items():
            for qubit, value in values.items():
                # resonator_spectroscopy / resonator_spectroscopy_flux / resonator_punchout_attenuation
                if par == "readout_frequency":
                    freq = int(value * 1e9)
                    mz = self.qubits[qubit].native_gates.MZ
                    mz.frequency = freq
                    if mz.if_frequency is not None:
                        mz.if_frequency = freq - self.get_lo_readout_frequency(qubit)
                    self.qubits[qubit].readout_frequency = freq

                # resonator_punchout_attenuation
                elif par == "readout_attenuation":
                    self.qubits[qubit].readout.attenuation = value

                # resonator_punchout_attenuation
                elif par == "bare_resonator_frequency":
                    freq = int(value * 1e9)
                    self.qubits[qubit].bare_resonator_frequency = freq

                # resonator_spectroscopy_flux / qubit_spectroscopy_flux
                elif par == "sweetspot":
                    sweetspot = float(value)
                    self.qubits[qubit].sweetspot = sweetspot
                    if self.qubits[qubit].flux is not None:
                        # set sweetspot as the flux offset (IS THIS NEEDED?)
                        self.qubits[qubit].flux.offset = sweetspot

                # qubit_spectroscopy / qubit_spectroscopy_flux / ramsey
                elif par == "drive_frequency":
                    if isinstance(value, tuple):  # TODO: remove this branching after error bars propagation
                        freq = int(value[0] * 1e9)
                    else:
                        freq = int(value * 1e9)
                    self.qubits[qubit].native_gates.RX.frequency = freq
                    self.qubits[qubit].drive_frequency = freq

                elif "amplitude" in par:
                    amplitude = float(value)
                    # resonator_spectroscopy
                    if par == "readout_amplitude" and not math.isnan(amplitude):
                        self.qubits[qubit].native_gates.MZ.amplitude = amplitude

                    # rabi_amplitude / flipping
                    if par == "drive_amplitude" or par == "amplitudes":
                        self.qubits[qubit].native_gates.RX.amplitude = amplitude

                # rabi_duration
                elif par == "drive_length":
                    self.qubits[qubit].native_gates.RX.duration = int(value)

                # ramsey
                elif par == "t2":
                    self.qubits[qubit].T2 = float(value)

                # spin_echo
                elif par == "t2_spin_echo":
                    self.qubits[qubit].T2_spin_echo = float(value)

                # t1
                elif par == "t1":
                    self.qubits[qubit].T1 = float(value)

                # classification
                elif par == "threshold":
                    self.qubits[qubit].threshold = float(value)

                # classification
                elif par == "iq_angle":
                    self.qubits[qubit].iq_angle = float(value)

                # classification
                elif par == "mean_gnd_states":
                    self.qubits[qubit].mean_gnd_states = [float(voltage) for voltage in value]

                # classification
                elif par == "mean_exc_states":
                    self.qubits[qubit].mean_exc_states = [float(voltage) for voltage in value]

                # drag pulse tunning
                elif "beta" in par:
                    rx = self.qubits[qubit].native_gates.RX
                    shape = rx.shape
                    rel_sigma = re.findall(r"[\d]+[.\d]+|[\d]*[.][\d]+|[\d]+", shape)[0]
                    rx.shape = f"Drag({rel_sigma}, {float(value)})"

                elif "length" in par:  # assume only drive length
                    self.qubits[qubit].native_gates.RX.duration = int(value)

                elif par == "classifiers_hpars":
                    self.qubits[qubit].classifiers_hpars = value

                else:
                    raise_error(ValueError, f"Unknown parameter {par} for qubit {qubit}")

    def connect(self):
        """Connect to all instruments."""
        if not self.is_connected:
            for instrument in self.instruments.values():
                try:
                    log.info(f"Connecting to instrument {instrument}.")
                    instrument.connect()
                except Exception as exception:
                    raise_error(
                        RuntimeError,
                        f"Cannot establish connection to {instrument} instruments. Error captured: '{exception}'",
                    )
        self.is_connected = True

    def setup(self):
        """Prepares instruments to execute experiments.

        Sets flux port offsets to the qubit sweetspots.
        """
        for instrument in self.instruments.values():
            instrument.setup()
        for qubit in self.qubits.values():
            if qubit.flux is not None and qubit.sweetspot != 0:
                qubit.flux.offset = qubit.sweetspot

    def start(self):
        """Starts all the instruments."""
        if self.is_connected:
            for instrument in self.instruments.values():
                instrument.start()

    def stop(self):
        """Starts all the instruments."""
        if self.is_connected:
            for instrument in self.instruments.values():
                instrument.stop()

    def disconnect(self):
        """Disconnects from instruments."""
        if self.is_connected:
            for instrument in self.instruments.values():
                instrument.disconnect()
        self.is_connected = False

    def _execute(self, method, sequences, options, **kwargs):
        """Executes the sequences on the controllers"""
        if options.nshots is None:
            options = replace(options, nshots=self.settings.nshots)

        if options.relaxation_time is None:
            options = replace(options, relaxation_time=self.settings.relaxation_time)

        duration = sum(seq.duration for seq in sequences) if isinstance(sequences, list) else sequences.duration
        time = (duration + options.relaxation_time) * options.nshots * 1e-9
        log.info(f"Minimal execution time (seq): {time}")

        result = {}
        for instrument in self.instruments.values():
            if isinstance(instrument, Controller):
                new_result = getattr(instrument, method)(self.qubits, sequences, options)
                if isinstance(new_result, dict):
                    result.update(new_result)
                elif new_result is not None:
                    # currently the result of QMSim is not a dict
                    result = new_result
        return result

    def execute_pulse_sequence(self, sequences: PulseSequence, options: ExecutionParameters, **kwargs):
        """
        Args:
            sequence (:class:`qibolab.pulses.PulseSequence`): Pulse sequences to execute.
            options (:class:`qibolab.platforms.platform.ExecutionParameters`): Object holding the execution options.
            **kwargs: May need them for something
        Returns:
            Readout results acquired by after execution.

        """

        return self._execute("play", sequences, options, **kwargs)

    def execute_pulse_sequences(self, sequences: List[PulseSequence], options: ExecutionParameters, **kwargs):
        """
        Args:
            sequence (List[:class:`qibolab.pulses.PulseSequence`]): Pulse sequences to execute.
            options (:class:`qibolab.platforms.platform.ExecutionParameters`): Object holding the execution options.
            **kwargs: May need them for something
        Returns:
            Readout results acquired by after execution.

        """
        return self._execute("play_sequences", sequences, options, **kwargs)

    def sweep(self, sequence: PulseSequence, options: ExecutionParameters, *sweepers: Sweeper):
        """Executes a pulse sequence for different values of sweeped parameters.

        Useful for performing chip characterization.

        Example:
            .. testcode::

                import numpy as np
                from qibolab.dummy import create_dummy
                from qibolab.sweeper import Sweeper, Parameter
                from qibolab.pulses import PulseSequence
                from qibolab.execution_parameters import ExecutionParameters


                platform = create_dummy()
                sequence = PulseSequence()
                parameter = Parameter.frequency
                pulse = platform.create_qubit_readout_pulse(qubit=0, start=0)
                sequence.add(pulse)
                parameter_range = np.random.randint(10, size=10)
                sweeper = Sweeper(parameter, parameter_range, [pulse])
                platform.sweep(sequence, ExecutionParameters(), sweeper)

        Returns:
            Readout results acquired by after execution.
        """
        if options.nshots is None:
            options = replace(options, nshots=self.settings.nshots)

        if options.relaxation_time is None:
            options = replace(options, relaxation_time=self.settings.relaxation_time)

        time = (sequence.duration + options.relaxation_time) * options.nshots * 1e-9
        for sweep in sweepers:
            time *= len(sweep.values)
        log.info(f"Minimal execution time (sweep): {time}")

        result = {}
        for instrument in self.instruments.values():
            if isinstance(instrument, Controller):
                new_result = instrument.sweep(self.qubits, sequence, options, *sweepers)
                if isinstance(new_result, dict):
                    result.update(new_result)
                elif new_result is not None:
                    # currently the result of QMSim is not a dict
                    result = new_result
        return result

    def __call__(self, sequence, options):
        return self.execute_pulse_sequence(sequence, options)

    def get_qubit(self, qubit):
        """Return the name of the physical qubit corresponding to a logical qubit.

        Temporary fix for the compiler to work for platforms where the qubits
        are not named as 0, 1, 2, ...
        """
        try:
            return self.qubits[qubit].name
        except KeyError:
            return list(self.qubits.keys())[qubit]

    def create_RX90_pulse(self, qubit, start=0, relative_phase=0):
        qubit = self.get_qubit(qubit)
        return self.qubits[qubit].native_gates.RX90.pulse(start, relative_phase)

    def create_RX_pulse(self, qubit, start=0, relative_phase=0):
        qubit = self.get_qubit(qubit)
        return self.qubits[qubit].native_gates.RX.pulse(start, relative_phase)

    def create_CZ_pulse_sequence(self, qubits, start=0):
        # Check in the settings if qubits[0]-qubits[1] is a key
        pair = tuple(sorted(self.get_qubit(q) for q in qubits))
        if pair not in self.pairs or self.pairs[pair].native_gates.CZ is None:
            raise_error(
                ValueError,
                f"Calibration for CZ gate between qubits {qubits[0]} and {qubits[1]} not found.",
            )
        return self.pairs[pair].native_gates.CZ.sequence(start)

    def create_MZ_pulse(self, qubit, start):
        qubit = self.get_qubit(qubit)
        return self.qubits[qubit].native_gates.MZ.pulse(start)

    def create_qubit_drive_pulse(self, qubit, start, duration, relative_phase=0):
        qubit = self.get_qubit(qubit)
        pulse = self.qubits[qubit].native_gates.RX.pulse(start, relative_phase)
        pulse.duration = duration
        return pulse

    def create_qubit_readout_pulse(self, qubit, start):
        qubit = self.get_qubit(qubit)
        return self.create_MZ_pulse(qubit, start)

    # TODO Remove RX90_drag_pulse and RX_drag_pulse, replace them with create_qubit_drive_pulse
    # TODO Add RY90 and RY pulses

    def create_RX90_drag_pulse(self, qubit, start, relative_phase=0, beta=None):
        qubit = self.get_qubit(qubit)
        pulse = self.qubits[qubit].native_gates.RX90.pulse(start, relative_phase)
        if beta is not None:
            pulse.shape = "Drag(5," + str(beta) + ")"
        return pulse

    def create_RX_drag_pulse(self, qubit, start, relative_phase=0, beta=None):
        qubit = self.get_qubit(qubit)
        pulse = self.qubits[qubit].native_gates.RX.pulse(start, relative_phase)
        if beta is not None:
            pulse.shape = "Drag(5," + str(beta) + ")"
        return pulse

    def set_lo_drive_frequency(self, qubit, freq):
        """Set frequency of the qubit drive local oscillator.

        Args:
            qubit (int): qubit whose local oscillator will be modified.
            freq (int): new value of the frequency in Hz.
        """
        self.qubits[qubit].drive.lo_frequency = freq

    def get_lo_drive_frequency(self, qubit):
        """Get frequency of the qubit drive local oscillator in Hz."""
        return self.qubits[qubit].drive.lo_frequency

    def set_lo_readout_frequency(self, qubit, freq):
        """Set frequency of the qubit drive local oscillator.

        Args:
            qubit (int): qubit whose local oscillator will be modified.
            freq (int): new value of the frequency in Hz.
        """
        self.qubits[qubit].readout.lo_frequency = freq

    def get_lo_readout_frequency(self, qubit):
        """Get frequency of the qubit readout local oscillator in Hz."""
        return self.qubits[qubit].readout.lo_frequency

    def set_lo_twpa_frequency(self, qubit, freq):
        """Set frequency of the local oscillator of the TWPA to which the qubit's feedline is connected to.

        Args:
            qubit (int): qubit whose local oscillator will be modified.
            freq (int): new value of the frequency in Hz.
        """
        self.qubits[qubit].twpa.lo_frequency = freq

    def get_lo_twpa_frequency(self, qubit):
        """Get frequency of the local oscillator of the TWPA to which the qubit's feedline is connected to in Hz."""
        return self.qubits[qubit].twpa.lo_frequency

    def set_lo_twpa_power(self, qubit, power):
        """Set power of the local oscillator of the TWPA to which the qubit's feedline is connected to.

        Args:
            qubit (int): qubit whose local oscillator will be modified.
            power (int): new value of the power in dBm.
        """
        self.qubits[qubit].twpa.lo_power = power

    def get_lo_twpa_power(self, qubit):
        """Get power of the local oscillator of the TWPA to which the qubit's feedline is connected to in dBm."""
        return self.qubits[qubit].twpa.lo_power

    def set_attenuation(self, qubit, att):
        """Set attenuation value. Usefeul for calibration routines such as punchout.

        Args:
            qubit (int): qubit whose attenuation will be modified.
            att (int): new value of the attenuation (dB).
        Returns:
            None
        """
        self.qubits[qubit].readout.attenuation = att

    def get_attenuation(self, qubit):
        """Get attenuation value. Usefeul for calibration routines such as punchout."""
        return self.qubits[qubit].readout.attenuation

    def set_gain(self, qubit, gain):
        """Set gain value. Usefeul for calibration routines such as Rabi oscillations.

        Args:
            qubit (int): qubit whose attenuation will be modified.
            gain (int): new value of the gain (dimensionless).
        Returns:
            None
        """
        raise_error(NotImplementedError, f"{self.name} does not support gain.")

    def get_gain(self, qubit):
        """Get gain value. Usefeul for calibration routines such as Rabi oscillations."""
        raise_error(NotImplementedError, f"{self.name} does not support gain.")

    def set_bias(self, qubit, bias):
        """Set bias value. Usefeul for calibration routines involving flux.

        Args:
            qubit (int): qubit whose attenuation will be modified.
            bias (int): new value of the bias (V).
        Returns:
            None
        """
        if self.qubits[qubit].flux is None:
            raise_error(NotImplementedError, f"{self.name} does not have flux.")
        self.qubits[qubit].flux.bias = bias

    def get_bias(self, qubit):
        """Get bias value. Usefeul for calibration routines involving flux."""
        return self.qubits[qubit].flux.bias



from qibolab.instruments.abstract import Instrument


class LocalOscillator(Instrument):
    """Abstraction for local oscillator instruments.

    Local oscillators are used to upconvert signals, when
    the controllers cannot send sufficiently high frequencies
    to address the qubits and resonators.
    They cannot be used to play or sweep pulses.
    """

    def __init__(self, name, address):
        super().__init__(name, address)
        self._power: float
        self._frequency: float

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, x):
        self._frequency = x

    @property
    def power(self):
        return self._power

    @power.setter
    def power(self, x):
        self._power = x

    def connect(self):
        self.is_connected = True

    def setup(self, power=None, frequency=None, **kwargs):
        if power is not None:
            self.power = power
        if frequency is not None:
            self.frequency = frequency

    def start(self):
        pass

    def stop(self):
        pass

    def disconnect(self):
        self.is_connected = False



import random

import networkx as nx
from qibo import gates
from qibo.config import raise_error
from qibo.models import Circuit

from qibolab.transpilers.abstract import Placer, Router, find_gates_qubits_pairs


class PlacementError(Exception):
    """Raise for an error in the qubit placement"""


def assert_placement(circuit: Circuit, layout: dict) -> bool:
    """Check if layout is correct and matches the number of qubits of the circuit.

    Args:
        circuit (qibo.models.Circuit): Circuit model to check.
        layout (dict): physical to logical qubit mapping.

    Raise PlacementError if the following conditions are not satisfied:
        - layout is written in the correct form.
        - layout matches the number of qubits in the circuit.
    """
    assert_mapping_consistency(layout)
    if circuit.nqubits > len(layout):
        raise PlacementError("Layout can't be used on circuit. The circuit requires more qubits.")
    if circuit.nqubits < len(layout):
        raise PlacementError("Layout can't be used on circuit. Ancillary extra qubits need to be added to the circuit.")


def assert_mapping_consistency(layout):
    """Check if layout is correct.

    Args:
        layout (dict): physical to logical qubit mapping.

    Raise PlacementError if layout is not written in the correct form.
    """
    values = sorted(layout.values())
    keys = list(layout)
    ref_keys = ["q" + str(i) for i in range(len(keys))]
    if keys != ref_keys:
        raise PlacementError("Some physical qubits in the layout may be missing or duplicated.")
    if values != list(range(len(values))):
        raise PlacementError("Some logical qubits in the layout may be missing or duplicated.")


class Trivial(Placer):
    """Place qubits according to the following simple notation: {'q0' : 0, 'q1' : 1, ..., 'qn' : n}.

    Attributes:
        connectivity (networkx.Graph): chip connectivity.
    """

    def __init__(self, connectivity: nx.Graph = None):
        self.connectivity = connectivity

    def __call__(self, circuit: Circuit):
        """Find the trivial placement for the circuit.

        Args:
            circuit (qibo.models.Circuit): circuit to be transpiled.

        Returns:
            (dict): physical to logical qubit mapping.
        """
        if self.connectivity is not None:
            if self.connectivity.number_of_nodes() != circuit.nqubits:
                raise PlacementError(
                    "The number of nodes of the connectivity graph must match the number of qubits in the circuit"
                )
        return dict(zip(list("q" + str(i) for i in range(circuit.nqubits)), range(circuit.nqubits)))


class Custom(Placer):
    """Define a custom initial qubit mapping.

    Attributes:
        map (list or dict): physical to logical qubit mapping,
        example [1,2,0] or {"q0":1, "q1":2, "q2":0} to assign the
        physical qubits 0;1;2 to the logical qubits 1;2;0 respectively.
        connectivity (networkx.Graph): chip connectivity.
    """

    def __init__(self, map, connectivity=None):
        self.connectivity = connectivity
        self.map = map

    def __call__(self, circuit=None):
        """Return the custom placement if it can be applied to the given circuit (if given).

        Args:
            circuit (qibo.models.Circuit): circuit to be transpiled.

        Returns:
            (dict): physical to logical qubit mapping.
        """
        if isinstance(self.map, dict):
            pass
        elif isinstance(self.map, list):
            self.map = dict(zip(list("q" + str(i) for i in range(len(self.map))), self.map))
        else:
            raise_error(TypeError, "Use dict or list to define mapping.")
        if circuit is not None:
            assert_placement(circuit, self.map)
        else:
            assert_mapping_consistency(self.map)
        return self.map


class Subgraph(Placer):
    """
    Subgraph isomorphism qubit placer, NP-complete it can take a long time
    for large circuits. This initialization method may fail for very short circuits.

    Attributes:
        connectivity (networkx.Graph): chip connectivity.
    """

    def __init__(self, connectivity):
        self.connectivity = connectivity

    def __call__(self, circuit: Circuit):
        """Find the initial layout of the given circuit using subgraph isomorphism.

        Args:
            circuit (qibo.models.Circuit): circuit to be transpiled.

        Returns:
            (dict): physical to logical qubit mapping.
        """
        gates_qubits_pairs = find_gates_qubits_pairs(circuit)
        if len(gates_qubits_pairs) < 3:
            raise_error(
                ValueError, "Circuit must contain at least two two qubit gates to implement subgraph placement."
            )
        circuit_subgraph = nx.Graph()
        circuit_subgraph.add_nodes_from(range(self.connectivity.number_of_nodes()))
        matcher = nx.algorithms.isomorphism.GraphMatcher(self.connectivity, circuit_subgraph)
        i = 0
        circuit_subgraph.add_edge(gates_qubits_pairs[i][0], gates_qubits_pairs[i][1])
        while matcher.subgraph_is_monomorphic() == True:
            result = matcher
            i += 1
            circuit_subgraph.add_edge(gates_qubits_pairs[i][0], gates_qubits_pairs[i][1])
            matcher = nx.algorithms.isomorphism.GraphMatcher(self.connectivity, circuit_subgraph)
            if (
                self.connectivity.number_of_edges() == circuit_subgraph.number_of_edges()
                or i == len(gates_qubits_pairs) - 1
            ):
                keys = list(result.mapping.keys())
                keys.sort()
                return {i: result.mapping[i] for i in keys}
        return dict(sorted(result.mapping.items()))


class Random(Placer):
    """
    Random initialization with greedy policy, let a maximum number of 2-qubit
    gates can be applied without introducing any SWAP gate.

    Attributes:
        connectivity (networkx.Graph): chip connectivity.
        samples (int): number of initial random layouts tested.
    """

    def __init__(self, connectivity, samples=100):
        self.connectivity = connectivity
        self.samples = samples

    def __call__(self, circuit):
        """Find an initial layout of the given circuit using random greedy algorithm.

        Args:
            circuit (qibo.models.Circuit): Circuit to be transpiled.

        Returns:
            (dict): physical to logical qubit mapping.
        """
        gates_qubits_pairs = find_gates_qubits_pairs(circuit)
        nodes = self.connectivity.number_of_nodes()
        keys = list(self.connectivity.nodes())
        final_mapping = dict(zip(keys, range(nodes)))
        final_graph = nx.relabel_nodes(self.connectivity, final_mapping)
        final_cost = self.cost(final_graph, gates_qubits_pairs)
        for _ in range(self.samples):
            mapping = dict(zip(keys, random.sample(range(nodes), nodes)))
            graph = nx.relabel_nodes(self.connectivity, mapping)
            cost = self.cost(graph, gates_qubits_pairs)
            if cost == 0:
                return mapping
            if cost < final_cost:
                final_graph = graph
                final_mapping = mapping
                final_cost = cost
        return final_mapping

    @staticmethod
    def cost(graph, gates_qubits_pairs):
        """
        Compute the cost associated to an initial layout as the lengh of the reduced circuit.

        Args:
            graph (networkx.Graph): current hardware qubit mapping.
            gates_qubits_pairs (list): circuit representation.

        Returns:
            (int): lengh of the reduced circuit.
        """
        for allowed, gate in enumerate(gates_qubits_pairs):
            if gate not in graph.edges():
                break
        return len(gates_qubits_pairs) - allowed


class ReverseTraversal(Placer):
    """
    Place qubits based on the algorithm proposed in https://doi.org/10.48550/arXiv.1809.02573.
    Works with ShortestPaths routing.

    Attributes:
        connectivity (networkx.Graph): chip connectivity.
        routing_algorithm (qibolab.transpilers.routing.Transpiler): routing algorithm.
        depth (int): number of two qubit gates considered before finding initial layout.
            if 'None' just one backward step will be implemented.
            If depth is greater than the number of two qubit gates in the circuit, the circuit will be routed more than once.
            Example: on a circuit with four two qubit gates A-B-C-D using depth = 6,
            the routing will be performed on the circuit C-D-D-C-B-A.
    """

    def __init__(self, connectivity: nx.Graph, routing_algorithm: Router, depth=None):
        self.connectivity = connectivity
        self.routing_algorithm = routing_algorithm
        self.depth = depth

    def __call__(self, circuit: Circuit):
        """Find the initial layout of the given circuit using Reverse Traversal placement.

        Args:
            circuit (qibo.models.Circuit): circuit to be transpiled.

        Returns:
            (dict): physical to logical qubit mapping.
        """

        initial_placer = Trivial(self.connectivity)
        initial_placement = initial_placer(circuit=circuit)
        new_circuit = self.assemble_circuit(circuit)
        final_placement = self.routing_step(initial_placement, new_circuit)
        return final_placement

    def assemble_circuit(self, circuit: Circuit):
        """Assemble a single circuit to apply Reverse Traversal placement based on depth.
        Example: for a circuit with four two qubit gates A-B-C-D using depth = 6,
        the function will return the circuit C-D-D-C-B-A.

        Args:
            circuit (qibo.models.Circuit): circuit to be transpiled.

        Returns:
            new_circuit (qibo.models.Circuit): assembled circuit to perform Reverse Traversal placement.
        """

        if self.depth is None:
            return circuit.invert()
        gates_qubits_pairs = find_gates_qubits_pairs(circuit)
        circuit_gates = len(gates_qubits_pairs)
        if circuit_gates == 0:
            raise ValueError("The circuit must contain at least a two qubit gate.")
        repetitions, remainder = divmod(self.depth, circuit_gates)
        assembled_gates_qubits_pairs = []
        for _ in range(repetitions):
            assembled_gates_qubits_pairs += gates_qubits_pairs[:]
            gates_qubits_pairs.reverse()
        assembled_gates_qubits_pairs += gates_qubits_pairs[0:remainder]
        new_circuit = Circuit(circuit.nqubits)
        for qubits in assembled_gates_qubits_pairs:
            # As only the connectivity is important here we can replace everything with CZ gates
            new_circuit.add(gates.CZ(qubits[0], qubits[1]))
        return new_circuit.invert()

    def routing_step(self, layout: dict, circuit: Circuit):
        """Perform routing of the circuit.

        Args:
            layout (dict): intial qubit layout.
            circuit (qibo.models.Circuit): circuit to be routed.
        """

        _, final_mapping = self.routing_algorithm(circuit, layout)
        return final_mapping
